from matplotlib.axes import Axes
from numpy import mean
from pandas import Series

from ux.plots.helpers import new_axes


def plot_task_learnability(task_times: dict, ax: Axes = None):
    """
    Plot the change in time on task over a number of trials,

    :param task_times: Dictionary of {trial name => list of task times}
    :param ax: Optional matplotlib axes to plot on.
    :rtype: Axes
    """
    means = Series({
        trial_name: mean(times)
        for trial_name, times in task_times.items()
    })
    ax = ax or new_axes()
    means.plot(kind='line', ax=ax, marker="s")
    ax.set_xlabel('Trials')
    ax.set_ylabel('Time-on-Task (Sec')
    ax.set_ylim(0, 80)

    return ax



from collections import defaultdict, OrderedDict
from datetime import datetime, timedelta
from pandas import notnull
from types import FunctionType
from typing import List, Callable, Set, Union

from ux.calcs.object_calcs.efficiency import lostness
from ux.calcs.object_calcs.task_success import unordered_task_completion_rate, ordered_task_completion_rate, \
    binary_task_success
from ux.calcs.object_calcs.utils import sequence_intersects_task
from ux.classes.wrappers.map_result import MapResult
from ux.custom_types import ActionMapper
from ux.interfaces.actions.i_action_template import IActionTemplate
from ux.interfaces.actions.i_user_action import IUserAction
from ux.interfaces.sequences.i_action_sequence import IActionSequence
from ux.interfaces.tasks.i_task import ITask
from ux.utils.actions import create_action_template_condition
from ux.utils.misc import get_method_name


class ActionSequence(IActionSequence):
    """
    Represents a sequence of UserActions taken by a User.
    """
    def __init__(self, user_actions: List[IUserAction] = None, meta: dict = None):
        """
        Create a new ActionSequence.

        :param user_actions: List of Actions to use to construct the ActionSequence.
        :param meta: Optional additional data to store with the ActionSequence.
        """
        self._user_actions = user_actions or []
        self._meta = meta
        self._action_templates = None
        self._location_ids = None

    @property
    def user_actions(self):
        """
        Return the list of UserActions in the ActionSequence.

        :rtype: List[IUserAction]
        """
        return self._user_actions

    @property
    def meta(self):
        """
        Return the dictionary of meta information added at construction time.

        :rtype: dict
        """
        return self._meta

    @property
    def start(self):
        """
        :rtype: datetime
        """
        return self[0].time_stamp

    @property
    def end(self):
        """
        :rtype: datetime
        """
        return self[-1].time_stamp

    @property
    def duration(self):
        """
        Return the total duration of the ActionSequence from the first Action to the last.

        :rtype: timedelta
        """
        start_time = self.user_actions[0].time_stamp
        end_time = self.user_actions[-1].time_stamp
        return end_time - start_time

    @property
    def user_id(self):
        """
        :rtype: str
        """
        return self[0].user_id

    @property
    def session_id(self):
        """
        :rtype: str
        """
        return self[0].session_id

    def action_templates(self):
        """
        Return a list of ActionTemplates derived from each of the UserActions taken.

        :rtype: List[IActionTemplate]
        """
        if self._action_templates is None:
            self._action_templates = [
                user_action.template()
                for user_action in self._user_actions
            ]
        return self._action_templates

    def action_template_set(self):
        """
        Return a set of ActionTemplates representing the UserActions in the ActionSequence.

        :rtype: Set[IActionTemplate]
        """
        return set(self.action_templates())

    def action_template_counts(self):
        """
        Return a count of each ActionTemplate representing one or more UserActions in the ActionSequence.

        :rtype: Dict[IActionTemplate, int]
        """
        counts = defaultdict(int)
        for template in self.action_templates():
            counts[template] += 1
        return dict(counts)

    def find_all(self, template: IActionTemplate):
        """
        Return a list of all the actions matching the given action template.
        Returns an empty list if the template is not matched.

        :param template: The ActionTemplate to match against.
        :type template: IActionTemplate
        :rtype: list[IUserAction]
        """
        return [action for action in self.user_actions
                if action.template() == template]

    def find_first(self, template: IActionTemplate):
        """
        Return the first action matching the given action template. Returns None if the template is not matched.

        :type template: IActionTemplate
        :rtype: IUserAction
        """
        occurrences = self.find_all(template)
        if len(occurrences):
            return occurrences[0]
        else:
            return None

    def find_last(self, template: IActionTemplate):
        """
        Return the first action matching the given action template. Returns None if the template is not matched.

        :type template: IActionTemplate
        :rtype: IUserAction
        """
        occurrences = self.find_all(template)
        if len(occurrences):
            return occurrences[-1]
        else:
            return None

    def map(self, mapper: Union[str, dict, list, ActionMapper]):
        """
        Apply a map function to every action in the Sequence and return the results.

        :param mapper: The method or methods to apply to each UserAction
        :rtype: MapResult
        """
        def map_items(item_mapper):
            if isinstance(item_mapper, str):
                # properties and methods
                if hasattr(IUserAction, item_mapper):
                    if callable(getattr(IUserAction, item_mapper)):
                        # methods
                        return [getattr(action, item_mapper)() for action in self._user_actions]
                    else:
                        # properties
                        return [getattr(action, item_mapper) for action in self._user_actions]
            elif isinstance(item_mapper, FunctionType):
                return [item_mapper(action) for action in self._user_actions]
            else:
                raise TypeError('item mappers must be FunctionType or str')

        if isinstance(mapper, str) or isinstance(mapper, FunctionType):
            results = OrderedDict([(get_method_name(mapper), map_items(mapper))])
        elif isinstance(mapper, dict):
            results = OrderedDict([
                (get_method_name(key), map_items(value))
                for key, value in mapper.items()
            ])
        elif isinstance(mapper, list):
            results = OrderedDict([
                (get_method_name(item), map_items(item))
                for item in mapper
            ])
        else:
            raise TypeError('mapper must be dict, str or FunctionType')

        return MapResult(results)

    def intersects_task(self, task: ITask):
        """
        Return True if the given Task has ActionTemplates that are equivalent to any Actions in the Sequence.

        :param task: Task to cross-reference Action Templates against.
        :rtype: bool
        """
        return sequence_intersects_task(action_sequence=self, task=task)

    def binary_task_success(self, task: ITask,
                            success_func: Callable[[ITask, IActionSequence], bool]):
        """
        Return True if success_func is met.

        :param task: The Task to assess success against.
        :param success_func: Callable to use to assess success.
        :rtype: bool
        """
        return binary_task_success(
            task=task, action_sequence=self,
            success_func=success_func
        )

    def lostness(self, task: ITask):
        """
        Return the lostness with respect to the given Task.

        :param task: The Task to calculate lostness against.
        :rtype: float
        """
        return lostness(task=task, action_sequence=self)

    def split(self, split, how: str = 'at', copy_meta: bool = False):
        """
        Split into a list of new ActionSequences after each `UserAction` where `condition` is met.

        :param split: Lambda function or Action Template to use to break the sequence.
        :type split: Union[ActionFilter, IActionTemplate]
        :param how: How to split the Sequence. One of `['before', 'after', 'at']`
        :param copy_meta: Whether to copy the `meta` dict into the new Sequences.
        :rtype: List[IActionSequence]
        """
        # find matching locations
        match_locs = []
        condition = create_action_template_condition(split)
        for a, action in enumerate(self.user_actions):
            if condition(action):
                match_locs.append(a)
        # split at right locations depending on the value of `how`
        if how == 'before':
            seq_starts = [m for m in match_locs]
            seq_ends = [m for m in match_locs if m != 0] + [len(self)]
            if match_locs[0] != 0:
                seq_starts.insert(0, 0)
            if match_locs[-1] == len(self):
                seq_ends = seq_ends[: -1]
        elif how == 'after':
            seq_starts = [0] + [m + 1 for m in match_locs]
            if seq_starts[-1] == len(self):
                seq_starts = seq_starts[: -1]
            seq_ends = [m + 1 for m in match_locs]
            if seq_ends[-1] != len(self):
                seq_ends.append(len(self))
        elif how == 'at':
            seq_starts = [m + 1 for m in match_locs if m != len(self) - 1]
            if match_locs[0] != 0:
                seq_starts.insert(0, 0)
            seq_ends = [m for m in match_locs if m != 0]
            if match_locs[-1] != len(self) - 1:
                seq_ends.append(len(self))
        else:
            raise ValueError("'how' must be set to one of ['before', 'after', 'at']")

        return [ActionSequence(
            user_actions=self.user_actions[start: end],
            meta=self._meta if copy_meta else None
        ) for start, end in zip(seq_starts, seq_ends)]

    def crop(self, start, end, how: str, copy_meta: bool = False):
        """
        Crop the sequence to start and end ActionTemplates or conditions.
        Returns None if both conditions are not found in order.

        :param start: The start of the subsequence to crop to.
        :param end: The end of the subsequence to crop to.
        :param how: 'first' or 'last'
        :param copy_meta: Whether to copy the `meta` dict into the new Sequence.
        :rtype: IActionSequence
        """
        start = create_action_template_condition(start)
        end = create_action_template_condition(end)
        a_start = None
        a_end = None
        if how == 'first':
            for a in range(len(self)):
                if start(self[a].template()):
                    a_start = a
                    break
            if a_start is not None and a_start < len(self) - 1:
                for a in range(a_start + 1, len(self)):
                    if end(self[a].template()):
                        a_end = a
                        break
        elif how == 'last':
            for a in range(len(self) - 1, -1, -1):
                if end(self[a].template()):
                    a_end = a
                    break
            if a_end is not None and a_end > 0:
                for a in range(a_end - 1, -1, -1):
                    if start(self[a].template()):
                        a_start = a
                        break
        else:
            raise ValueError("'how' must be one of 'first' or 'last'")
        if None in (a_start, a_end):
            return None
        else:
            return ActionSequence(
                user_actions=self.user_actions[a_start: a_end + 1],
                meta=self._meta if copy_meta else None
            )

    def unordered_completion_rate(self, task: ITask):
        """
        Calculate the unordered completion rate of the given Task from the Actions in the Sequence.

        :param task: The Task to cross-reference UserActions against.
        :rtype: float
        """
        return unordered_task_completion_rate(task, self)

    def ordered_completion_rate(self, task: ITask):
        """
        Calculate the ordered completion rate of the given Task from the Actions in the Sequence.

        :param task: The Task to cross-reference UserActions against.
        :rtype: float
        """
        return ordered_task_completion_rate(task, self)

    def location_ids(self):
        """
        Return a set of the ids of the unique locations visited in the sequence.

        :rtype: Set[str]
        """
        if self._location_ids is None:
            location_ids = set()
            for action in self._user_actions:
                location_ids.add(action.source_id)
                if action.target_id:
                    location_ids.add(action.target_id)
            self._location_ids = location_ids
        return self._location_ids

    def contains_location_id(self, location_id: str):
        """
        Determine whether the location was visited in the sequence.

        :rtype: bool
        """
        return location_id in self.location_ids()

    def action_types(self):
        """
        Return a set of the unique action types carried out in the sequence.

        :rtype: Set[str]
        """
        return set([action.action_type for action in self.user_actions])

    def back_click_rates(self):
        """
        :rtype: Dict[IActionTemplate, float]
        """
        rates = {}
        template_list = self.action_templates()
        for template in self.action_template_set():
            forwards = template_list.count(template)
            backwards = template_list.count(template.reversed())
            rate = backwards / forwards
            if rate <= 1:
                rates[template] = rate
        return rates

    def dwell_times(self, sum_by_location: bool):
        """
        Return the amount of time spent by the user at each location.

        :param sum_by_location: Whether to sum the durations of time spent at each location or keep as a list.
        :rtype: Dict[str, Union[timedelta, List[timedelta]]]
        """
        if sum_by_location:
            dwell_times = defaultdict(timedelta)
            for a in range(1, len(self)):
                prev_action = self[a - 1]
                next_action = self[a]
                if notnull(prev_action.target_id) and prev_action.target_id != '':
                    location = prev_action.target_id
                else:
                    location = prev_action.source_id
                dwell_times[location] += next_action.time_stamp - prev_action.time_stamp
        else:
            dwell_times = defaultdict(list)
            for a in range(1, len(self)):
                prev_action = self[a - 1]
                next_action = self[a]
                if notnull(prev_action.target_id) and prev_action.target_id != '':
                    location = prev_action.target_id
                else:
                    location = prev_action.source_id
                dwell_times[location].append(
                    next_action.time_stamp - prev_action.time_stamp
                )
        return dwell_times

    def __getitem__(self, item):
        """
        :rtype: IUserAction
        """
        return self._user_actions[item]

    def __repr__(self):
        """
        :rtype: str
        """
        return 'ActionSequence([{}])'.format(
            len(self._user_actions)
        )

    def __len__(self):
        """
        :rtype: int
        """
        return len(self._user_actions)

    def __contains__(self, item):
        """
        :rtype: bool
        """
        if isinstance(item, IUserAction):
            return item in self._user_actions
        elif isinstance(item, IActionTemplate):
            return item in self.action_templates()
        else:
            raise TypeError('item must be IUserAction or IActionTemplate')

    def __iter__(self):
        """
        :rtype: IUserAction
        """
        return self._user_actions.__iter__()



from ux.interfaces.i_location import ILocation


class Location(ILocation):
    """
    Represents a Location where an Action could be taken.
    """
    def __init__(self, location_id: str):
        """
        Create a new Location.

        :param location_id: The id of the Location.
        """
        self._location_id = location_id

    @property
    def location_id(self):
        """
        Return the id of the Location.

        :rtype: str
        """
        return self._location_id

    def __repr__(self):

        return 'Location({})'.format(self._location_id)

    def __gt__(self, other):
        """
        :type other: Location
        """
        return self._location_id > other._location_id

    def __lt__(self, other):
        """
        :type other: Location
        """
        return self._location_id < other._location_id



from typing import Iterable

from ux.classes.stats.confidence_interval import ConfidenceInterval


class SuccessRate(object):
    """
    Represents a Task Success Rate.
    """
    def __init__(self, mean: float, confidence_interval: ConfidenceInterval,
                 name: str = None):
        """
        Create a new SuccessRate.

        :param mean: The mean value of the results.
        :param confidence_interval: The confidence interval of the results.
        :param name: The name for the SuccessRate.
        """
        self.mean = mean
        self.confidence_interval = confidence_interval
        self.name = name or ''

    def __str__(self):

        return 'Mean: {:.2f}, CI: {:.2f}'.format(
            self.mean, self.confidence_interval.width
        )

    @staticmethod
    def means(success_rates):
        """
        Return a list of the mean value of each SuccessRate.

        :type success_rates: Iterable[SuccessRate]
        :rtype: list
        """
        return [
            success_rate.mean for success_rate in success_rates
        ]

    @property
    def errors(self):
        """
        Return the lower and upper error of the confidence interval bounds from the mean.

        :rtype: Tuple[float, float]
        """
        return (
            self.mean - self.confidence_interval.lower,
            self.confidence_interval.upper - self.mean
        )



from collections import defaultdict
from typing import List

from ux.interfaces.sequences.i_action_sequence import IActionSequence
from ux.interfaces.tasks.i_task import ITask


def transition_counts(task: ITask, action_sequences: List[IActionSequence]):
    """
    Count the transitions between actions.

    :param task: The task defining the actions that should  be counted.
    :param action_sequences: List of action sequences to count over.
    :return: Dictionary of {(from, to) => count}
    :rtype: Dict[Tuple[IActionTemplate, IActionTemplate], int]
    """
    transitions = defaultdict(int)
    task_action_set = set(task.action_templates)
    # count transitions
    for sequence in action_sequences:
        if not task.intersects_sequence(sequence):
            continue
        for a in range(len(sequence) - 1):
            from_action = sequence.user_actions[a].template()
            to_action = sequence.user_actions[a + 1].template()
            if {from_action, to_action}.issubset(task_action_set):
                transitions[(from_action, to_action)] += 1
    return transitions


def transition_probabilities(task: ITask, action_sequences: List[IActionSequence]):
    """
    Calculate the conditional probability of a first-order Markov Chain between actions.

    :param task: The task defining the actions that should  be counted.
    :param action_sequences: List of action sequences to count over.
    :return: Dictionary of {(from, to) => p(to|from)}
    :rtype: Dict[Tuple[ITask, ITask], float]
    """
    transitions = defaultdict(int)
    source_counts = defaultdict(int)
    task_action_set = set(task.action_templates)
    # count transitions and occurrences
    for sequence in action_sequences:
        if not task.intersects_sequence(sequence):
            continue
        for a in range(len(sequence) - 1):
            from_action = sequence.user_actions[a].template()
            to_action = sequence.user_actions[a + 1].template()
            if {from_action, to_action}.issubset(task_action_set):
                transitions[(from_action, to_action)] += 1
                source_counts[from_action] += 1
    # calculate conditional probabilities
    transition_probs = {}
    for (from_state, to_state), count in transitions.items():
        transition_probs[(from_state, to_state)] = (
                transitions[(from_state, to_state)] /
                source_counts[from_state]
        )

    return transition_probs



from collections import OrderedDict
from matplotlib.axes import Axes
from numpy import array, ndarray, histogram
from pandas import DataFrame, Series
from typing import Union, Iterable

from ux.calcs.basic_calcs.task_success import binary_task_success_rate
from ux.plots.helpers import new_axes, get_hist_index


def plot_task_success_rates(task_results: dict, ax: Axes = None):
    """
    Plot a bar plot of success rates on several tasks.

    :param task_results: Dictionary {task_name => task results list}.
    :param ax: Optional matplotlib axes.
    :rtype: Axes
    """
    # calculate plot data
    # TODO: debug confidence interval to match the book
    success_rates = {
        task_name: binary_task_success_rate(task_results[task_name])
        for task_name in task_results.keys()
    }
    means = Series({
        task_name: success_rates[task_name][0]
        for task_name in success_rates.keys()
    })
    confidence_intervals = array([
        success_rates[task_name][1]
        for task_name in means.index
    ])
    errors = confidence_intervals.copy()
    errors[:, 0] = means - errors[:, 0]
    errors[:, 1] = errors[:, 1] - means
    errors = errors.reshape(1, 2, means.shape[0])
    ax = ax or new_axes()
    means *= 100
    errors *= 100
    means.plot(kind='bar', ax=ax, yerr=errors, capsize=5)
    ax.set_xlabel('Task')
    ax.set_ylabel('% Successful')
    ax.set_ylim(0, 105)
    return ax


def plot_task_success_frequencies(
        success_rates, bins: Union[list, range] = None,
        ax: Axes = None
):
    """
    Plot frequencies of success rates for one or many scenarios.

    :param success_rates: List of success rates for one scenario, or dict(name->success_rates) for many.
    :type success_rates: Union[List[float], Dict[str, List[float]]]
    :param bins: Bins for the histogram calculation.
    :param ax: Optional matplotlib axes.
    :rtype: Axes
    """
    bins = bins or range(0, 110, 10)
    if type(success_rates) in [ndarray, list]:
        means = Series(
            data=success_rates,
            name='Success Rate'
        )
        means *= 100
        hist = Series(histogram(means, bins=bins)[0])
        hist.index = get_hist_index(bins)
        ax = ax or new_axes()
        hist.plot(kind='bar', ax=ax)
        ax.set_xlabel('Success Rate')
        ax.set_ylabel('Frequency')
        return ax
    elif type(success_rates) in (dict, OrderedDict):
        means_dict = dict()
        for name in success_rates.keys():
            means = array(success_rates[name])
            means *= 100
            means_dict[name] = histogram(means, bins=bins)[0]
        df = DataFrame(means_dict)
        df.index = get_hist_index(bins)
        ax = ax or new_axes()
        df.plot(kind='bar', ax=ax)
        ax.set_xlabel('Success Rate')
        ax.set_ylabel('Frequency')
        return ax
    else:
        raise TypeError


def plot_task_completion_rates(completion_rates, bins: Union[list, range] = None,
                               ax: Axes = None):

    """
    Plot rate of task completion for a given task.

    :param completion_rates: Completion rates for each participant (0 - 1)
    :type completion_rates: Union[List[float], Dict[str, List[float]]]
    :param bins: Bins for the histogram calculation.
    :param ax: Optional matplotlib axes.
    :rtype: Axes
    """
    ax = ax or new_axes()
    bins = bins or range(0, 110, 10)
    hist_index = get_hist_index(bins)

    def get_hist_series(rates: list):
        rates = Series(rates)
        rates *= 100
        hist = Series(histogram(rates, bins=bins)[0])
        hist.index = hist_index
        return hist

    if type(completion_rates) is list:
        data = get_hist_series(completion_rates)
    elif type(completion_rates) in (dict, OrderedDict):
        data = DataFrame({
            k: get_hist_series(rates=v)
            for k, v in completion_rates.items()
        })
    else:
        raise TypeError

    data.plot(kind='bar', ax=ax)
    ax.set_xlabel('Completion Rate')
    ax.set_ylabel('Frequency')


def plot_task_success_level(task_results: dict,
                            bar_order: Iterable[str] = None,  colors: Iterable[str] = None,
                            ax: Axes = None):
    """
    Plot a stacked bar chart of levels of success over different tasks.

    :param task_results: Dictionary {task name => {LoS => num participants})
    :param ax: Optional matplotlib axes.
    :rtype: Axes
    """
    df = DataFrame.from_dict(task_results, orient='index')
    df_percent = df.apply(lambda x: 100 * x / x.sum(), axis=1)
    ax = ax or new_axes()
    if bar_order is not None:
        df_percent = df_percent[bar_order]
    df_percent.plot(kind='bar', stacked=True, ax=ax, color=colors)
    ax.set_xlabel('Task')
    ax.set_ylabel('% of Participants')
    return ax



from matplotlib.axes import Axes
from numpy import array, mean
from pandas import Series, DataFrame

from ux.calcs.basic_calcs.stats import exponential_confidence_interval
from ux.plots.helpers import new_axes


def plot_task_completion_times(task_times: dict, confidence: float = 0.95, ax: Axes = None):
    """
    Plot the average time taken and confidence interval for each task.

    :param task_times: Dictionary of {task name => list of task times}
    :param confidence: The confidence interval to use for error bars (0 - 1)
    :param ax: Optional matplotlib axes to plot on.
    :rtype: Axes
    """
    means = Series({
        task_name: mean(times)
        for task_name, times in task_times.items()
    })
    errors = array([
        exponential_confidence_interval(task_times[task_name], confidence=confidence)
        for task_name in means.index
    ]).reshape(1, 2, means.shape[0])
    ax = ax or new_axes()
    means.plot(kind='bar', ax=ax, yerr=errors, capsize=5)
    ax.set_xlabel('Task')
    ax.set_ylabel('Time (Sec) to Complete Task')

    return ax


def plot_task_completion_under_threshold(participant_times: dict, threshold: int = 60, ax: Axes = None):
    """
    Plot the percentage of participants who completed each task within the time threshold.

    :param participant_times: Dictionary mapping  {task name => {participant => time on task}}
    :param threshold: Time threshold to use to measure task completion success.
    :param ax: Optional matplotlib axes to plot on.
    :rtype: Axes
    """
    df = DataFrame.from_dict(participant_times)
    df_below_threshold = df.le(threshold).mean()
    ax = ax or new_axes()
    df_below_threshold.plot(kind='bar', ax=ax)
    ax.set_xlabel('Task')
    ax.set_ylabel('Participants')
    percents = ax.get_yticks()
    ax.set_yticklabels(['{:,.0%}'.format(percent) for percent in percents])
    return ax



class ITask(object):

    @property
    def name(self):
        """
        :rtype: str
        """
        raise NotImplementedError

    @property
    def action_templates(self):
        """
        :rtype: List[IActionTemplate]
        """
        raise NotImplementedError

    def unordered_completion_rate(self, action_sequence):
        """
        :type action_sequence: IActionSequence
        :rtype: float
        """
        raise NotImplementedError

    def ordered_completion_rate(self, action_sequence):
        """
        :type action_sequence: IActionSequence
        :rtype: float
        """
        raise NotImplementedError

    def intersects_sequence(self, action_sequence):
        """
        :type action_sequence: IActionSequence
        :rtype: bool
        """
        raise NotImplementedError

    def __getitem__(self, item):
        """
        :rtype: IActionTemplate
        """
        raise NotImplementedError

    def __len__(self):
        """
        :rtype: int
        """
        raise NotImplementedError

    def __contains__(self, item):
        """
        :rtype: bool
        """
        raise NotImplementedError

    def __iter__(self):
        """
        :rtype: IActionTemplate
        """
        raise NotImplementedError



from datetime import datetime

from ux.interfaces.i_session import ISession


class Session(ISession):
    """
    Represents a single User Session.
    """
    def __init__(self, session_id: str, user_id: str,
                 start_time: datetime, end_time: datetime):
        """
        Create a new Session.

        :param session_id: The id of the Session.
        :param user_id: The id of the User whose Session this is.
        :param start_time: The start date-time of the Session.
        :param end_time: The end date-time of the Session.
        """
        self._session_id = session_id
        self._user_id = user_id or None
        self._start_time = start_time
        self._end_time = end_time

    @property
    def session_id(self):
        """
        Return the id of the Session.

        :rtype: str
        """
        return self._session_id

    @property
    def user_id(self):
        """
        Return the id of the Session's User.

        :rtype: str
        """
        return self._user_id

    @property
    def start_time(self):
        """
        Return the start date-time of the Session

        :rtype: datetime
        """
        return self._start_time

    @property
    def end_time(self):
        """
        Return the end date-time of the Session

        :rtype: datetime
        """
        return self._end_time

    def __repr__(self):

        return 'Session({})'.format(self._session_id)



from math import sqrt


def lostness(unique: int, total: int, optimum: int):
    """
    Calculate the `lostness` metric for the given pages.

    :param total: The total number of locations visited while performing the task, counting revisits to the same page.
    :param unique: The number of different locations visited while performing the task.
    :param optimum: The minimum (optimum) number of locations that must be visited to accomplish the task.
    :rtype: float
    """
    return sqrt(
        (unique / total - 1) ** 2 +
        (optimum / unique - 1) ** 2
    )



from collections import defaultdict
from pandas import Series, pivot_table, DataFrame, notnull
from typing import List

from ux.interfaces.actions.i_user_action import IUserAction
from ux.interfaces.sequences.i_action_sequence import IActionSequence


def count_action_transitions(action_sequences):
    """
    Count the transitions from each action to each other action in the given sequences.

    :param action_sequences: List of IActionSequence to count transitions in.
    :type action_sequences: List[IActionSequence]
    :return: Dictionary of {(from, to) => count}
    :rtype: dict[tuple[IActionTemplate, IActionTemplate], int]
    """
    transitions = defaultdict(int)
    # count transitions
    for sequence in action_sequences:
        for a in range(len(sequence) - 1):
            from_action = sequence.user_actions[a].template()
            to_action = sequence.user_actions[a + 1].template()
            transitions[(from_action, to_action)] += 1
    return transitions


def count_location_transitions(action_sequences):
    """
    Count the transitions from each location to each other location in actions in the given sequences.

    :param action_sequences: List of IActionSequence to count transitions in.
    :type action_sequences: List[IActionSequence]
    :rtype: dict[tuple[str, str], int]
    """
    transitions = defaultdict(int)
    # count transitions
    for sequence in action_sequences:
        for action in sequence.user_actions:
            source = action.source_id
            target = action.target_id
            if notnull(source) and notnull(target):
                transitions[(source, target)] += 1
    return transitions


def create_transition_table(transitions: dict, get_name=None, exclude=None):
    """
    Create a DataFrame of transitions with columns ['from', 'to', 'count']

    :param transitions: Dictionary of transitions and their counts or probabilities.
    :type transitions: Dict[Tuple[object, object], Union[float, int]]
    :param get_name: Optional lambda function to call to convert states to labels.
    :type get_name: Callable[[IUserAction], str]
    :param exclude: Optional list of names to exclude from the table.
    :type exclude: Union[str, List[str]]
    :rtype: DataFrame
    """
    def get_source_id(action: IUserAction):
        return action.source_id

    if get_name is None:
        first_from = list(transitions.items())[0][0]
        if isinstance(first_from, IUserAction):
            get_name = get_source_id

    transitions = Series(transitions).reset_index()
    transitions.columns = ['from', 'to', 'count']
    transitions = transitions.loc[
        (transitions['from'].notnull()) &
        (transitions['to'].notnull())
    ]
    if get_name is not None:
        transitions['from'] = transitions['from'].map(get_name)
        transitions['to'] = transitions['to'].map(get_name)
    if exclude is not None:
        if type(exclude) is str:
            exclude = [exclude]
        transitions = transitions.loc[
            (~transitions['from'].isin(exclude)) &
            (~transitions['to'].isin(exclude))
        ]
    return transitions


def create_transition_matrix(transitions: dict, get_name=None,
                             order_by='from', exclude=None):
    """
    Create a transition matrix from a dictionary of transition counts.

    :param transitions: Dictionary of transitions and their counts or probabilities.
    :type transitions: Dict[Tuple[object, object], Union[float, int]]
    :param get_name: Optional lambda function to call to convert states to labels.
    :type get_name: Callable[[IUserAction], str]
    :param order_by: Order labels by descending count of `from` or `to`, or pass a list to set order explicitly.
    :type order_by: Union[str, List[str]]
    :param exclude: Optional list of labels to exclude from the plots.
    :type exclude: Union[str, List[str]]
    :rtype: DataFrame
    """
    transitions = create_transition_table(transitions=transitions, get_name=get_name, exclude=exclude)
    if type(order_by) is str:
        order_names = transitions.groupby(order_by).sum()['count'].sort_values(ascending=False).index.tolist()
    else:
        order_names = order_by
    matrix = pivot_table(
        data=transitions, values='count',
        index='to', columns='from'
    )
    matrix = matrix.loc[filter(lambda n: n in matrix.index, order_names)]
    matrix = matrix[filter(lambda n: n in matrix.columns, order_names)]
    return matrix


def find_most_probable_sequence(transitions, get_name=None, exclude=None, start_at: str = None):
    """
    Find the most probable transition sequence.

    :param transitions: Dictionary of transitions and their counts or probabilities.
    :type transitions: Dict[Tuple[object, object], Union[float, int]]
    :param get_name: Optional lambda function to call to convert states to labels.
    :type get_name: Callable[[IUserAction], str]
    :param exclude: Optional list of names to exclude from the sequence.
    :type exclude: Union[str, List[str]]
    :param start_at: Optional name of start state. Leave as None to use most common state.
    :rtype: List[str]
    """
    transitions = create_transition_table(transitions=transitions, get_name=get_name, exclude=exclude)
    # find most frequent from point
    if start_at is not None:
        current_name = start_at
    else:
        current_name = transitions.groupby('from').sum()['count'].sort_values(ascending=False).index[0]
    sequence = [current_name]
    # iteratively find most frequent transition point from the current point
    found = True
    while found:
        transitions = transitions.loc[transitions['to'] != current_name]
        if current_name in transitions['from'].tolist():
            current_name = transitions.loc[
                transitions['from'] == current_name
            ].sort_values('count', ascending=False).iloc[0]['to']
            if current_name not in sequence:
                sequence.append(current_name)
            else:
                found = False
        else:
            found = False
    return sequence



import unittest
from unittest import TestCase

from ux.calcs.object_calcs.time_on_task import task_extents
from ux.classes.sequences.action_sequence import ActionSequence
from ux.classes.actions.action_template import ActionTemplate
from ux.classes.tasks.task import Task
from ux.classes.actions.user_action import UserAction


class TestTimeOnTask(TestCase):

    def setUp(self):

        self.test_task = Task(
            name='TestTask',
            action_templates=[
                ActionTemplate(action_type='action', source_id='location-1'),
                ActionTemplate(action_type='action', source_id='location-2'),
                ActionTemplate(action_type='action', source_id='location-3')
            ]
        )

        self.test_actions = ActionSequence(
            user_actions=[
                UserAction(action_id='1', action_type='action', source_id='location-A',
                           time_stamp=None, user_id=None, session_id=None),
                UserAction(action_id='2', action_type='action', source_id='location-1',
                           time_stamp=None, user_id=None, session_id=None),
                UserAction(action_id='3', action_type='action', source_id='location-2',
                           time_stamp=None, user_id=None, session_id=None),
                UserAction(action_id='5', action_type='action', source_id='location-B',
                           time_stamp=None, user_id=None, session_id=None),
                UserAction(action_id='4', action_type='action', source_id='location-3',
                           time_stamp=None, user_id=None, session_id=None),
                UserAction(action_id='5', action_type='action', source_id='location-C',
                           time_stamp=None, user_id=None, session_id=None)
            ]
        )

    def test_task_extents(self):

        self.assertEqual(
            (1, 4),
            task_extents(self.test_actions, self.test_task)
        )


if __name__ == '__main__':

    unittest.main()



from typing import List

from ux.calcs.basic_calcs.task_success import binary_task_success_rate


class TaskResult(object):
    """
    Represents the Result of a Task.
    """
    def __init__(self, value, meta: dict = None):
        """
        Create a new Task Result.

        :param value: The numeric value of the result.
        :param meta: Additional metadata about the TaskResult.
        """
        self.value = value
        self._meta = meta

    @staticmethod
    def binary_task_success_rate(results,
                                 alpha: float = 0.05, method: str = 'normal'):
        """
        Return the binary success rate of a number of TaskResults.

        :type results: List[TaskResult]
        :param results: list of pass (1/True) / fail (0/False) results
        :param alpha: significance level
        :param method: method to use for confidence interval
        :rtype: SuccessRate
        :return: (mean, confidence_interval)
        """
        result_values = [result.value for result in results]
        return binary_task_success_rate(
            results=result_values, alpha=alpha,
            method=method
        )


if __name__ == '__main__':

    r1 = (0.5, {'task': 'task 1', 'user': 'user 1'})
    r2 = (0.6, {'task': 'task 1', 'user': 'user 2'})
    r3 = (0.7, {'task': 'task 2', 'user': 'user 1'})
    r4 = (0.8, {'task': 'task 2', 'user': 'user 2'})

    tr1 = TaskResult(*r1)
    tr2 = TaskResult(*r2)
    tr3 = TaskResult(*r3)
    tr4 = TaskResult(*r4)



from collections import OrderedDict
from datetime import date, timedelta, datetime
from typing import List, Tuple

from ux.interfaces.sequences.i_action_sequence import IActionSequence
from ux.utils.dates import monday_on_or_before, date_to_datetime


def _get_sequence_start_end_dates(sequences: List[IActionSequence]):
    """
    Find the start and end dates of the sequences using the first UserAction in each sequence.

    :param sequences: List of IActionSequences to find dates from.
    :rtype: Tuple[date, date]
    """
    start_date = None
    end_date = None
    for sequence in sequences:
        sequence_date = sequence.user_actions[0].time_stamp.date()
        if start_date is None or sequence_date < start_date:
            start_date = sequence_date
        if end_date is None or sequence_date > end_date:
            end_date = sequence_date
    return start_date, end_date


def _get_sequence_start_end_date_times(sequences: List[IActionSequence]):
    """
    Find the start and end date-times of the sequences using the first UserAction in each sequence.

    :param sequences: List of IActionSequences to find dates from.
    :rtype: Tuple[datetime, datetime]
    """
    start_date_time = None
    end_date_time = None
    for sequence in sequences:
        sequence_date_time = sequence.user_actions[0].time_stamp
        if start_date_time is None or sequence_date_time < start_date_time:
            start_date_time = sequence_date_time
        if end_date_time is None or sequence_date_time > end_date_time:
            end_date_time = sequence_date_time
    return start_date_time, end_date_time


def split_sequences_by_hour(sequences: List[IActionSequence],
                            start_date_time: datetime = None, end_date_time: datetime = None):
    """
    Split a list of ActionSequences into an OrderedDict mapping each hour to a new list.

    Hours without any Sequences will contain an empty list.

    :param sequences: Original list of sequences to split by day.
    :param start_date_time: Optional first date-time to use to split the sequences.
    :param end_date_time: Optional last date-time to use.
    :rtype: OrderedDict[date, List[IActionSequence]]
    """
    if not sequences:
        return OrderedDict()
    # get start and end dates
    min_date, max_date = _get_sequence_start_end_date_times(sequences)
    if start_date_time is None:
        start_date_time = datetime(min_date.year, min_date.month, min_date.day,
                                   min_date.hour, 0, 0)
    if end_date_time is None:
        end_date_time = datetime(max_date.year, max_date.month, max_date.day,
                                 max_date.hour, 0, 0)
    # build the lists of sequences
    current_date_time = start_date_time
    sequence_dict = OrderedDict()
    while current_date_time <= end_date_time:
        date_sequences = [
            sequence for sequence in sequences
            if sequence.user_actions[0].time_stamp.date() == current_date_time.date() and
               sequence.user_actions[0].time_stamp.hour == current_date_time.hour
        ]
        sequence_dict[current_date_time] = date_sequences
        current_date_time = current_date_time + timedelta(hours=1)

    return sequence_dict


def split_sequences_by_day(sequences: List[IActionSequence],
                           start_date: date = None, end_date: date = None):
    """
    Split a list of ActionSequences into an OrderedDict mapping each date to a new list.

    Dates without any Sequences will contain an empty list.

    :param sequences: Original list of sequences to split by day.
    :param start_date: Optional first date to use to split the sequences.
    :param end_date: Optional last date to use.
    :rtype: OrderedDict[date, List[IActionSequence]]
    """
    if not sequences:
        return OrderedDict()
    # get start and end dates
    min_date, max_date = _get_sequence_start_end_dates(sequences)
    if start_date is None:
        start_date = min_date
    if end_date is None:
        end_date = max_date
    # build the lists of sequences
    current_date = start_date
    sequence_dict = OrderedDict()
    while current_date <= end_date:
        date_sequences = [
            sequence for sequence in sequences
            if sequence.user_actions[0].time_stamp.date() == current_date
        ]
        sequence_dict[current_date] = date_sequences
        current_date = current_date + timedelta(days=1)

    return sequence_dict


def split_sequences_by_week(sequences: List[IActionSequence],
                            start_date: date = None, end_date: date = None):
    """
    Split a list of ActionSequences into an OrderedDict mapping each week's start date to a new list.

    Dates without any Sequences will contain an empty list.

    :param sequences: Original list of sequences to split by day.
    :param start_date: Optional first date to use to split the sequences.
    :param end_date: Optional last date to use.
    :rtype: OrderedDict[date, List[IActionSequence]]
    """
    if not sequences:
        return OrderedDict()
    # get start and end dates
    min_date, max_date = _get_sequence_start_end_dates(sequences)
    if start_date is None:
        start_date = min_date
    if end_date is None:
        end_date = max_date
    start_date = monday_on_or_before(start_date)
    end_date = monday_on_or_before(end_date)
    # build the lists of sequences
    current_date = start_date
    sequence_dict = OrderedDict()
    while current_date <= end_date:
        first_date_time = date_to_datetime(current_date)
        last_date_time = date_to_datetime(current_date + timedelta(days=7))
        date_sequences = [
            sequence for sequence in sequences
            if (
                    first_date_time <=
                    sequence.user_actions[0].time_stamp <
                    last_date_time
            )
        ]
        sequence_dict[current_date] = date_sequences
        current_date = current_date + timedelta(days=7)

    return sequence_dict


def split_sequences_by_month(sequences: List[IActionSequence],
                             start_date: date = None, end_date: date = None):
    """
    Split a list of ActionSequences into an OrderedDict mapping each month's start date to a new list.

    Dates without any Sequences will contain an empty list.

    :param sequences: Original list of sequences to split by day.
    :param start_date: Optional first date to use to split the sequences.
    :param end_date: Optional last date to use.
    :rtype: OrderedDict[date, List[IActionSequence]]
    """
    if not sequences:
        return OrderedDict()
    # get start and end dates
    min_date, max_date = _get_sequence_start_end_dates(sequences)
    if start_date is None:
        start_date = min_date
    if end_date is None:
        end_date = max_date
    start_date = date(start_date.year, start_date.month, 1)
    end_date = date(end_date.year, end_date.month, 1)
    # build the lists of sequences
    current_date = start_date
    sequence_dict = OrderedDict()
    while current_date <= end_date:
        first_date = current_date
        last_date = (
            date(first_date.year, first_date.month + 1, 1)
            if first_date.month < 12
            else date(first_date.year + 1, 1, 1)
        )
        date_sequences = [
            sequence for sequence in sequences
            if first_date <= sequence.user_actions[0].time_stamp.date() < last_date
        ]
        sequence_dict[current_date] = date_sequences
        current_date = last_date

    return sequence_dict



from datetime import datetime

from ux.interfaces.actions.i_action_template import IActionTemplate


class IUserAction(object):

    @property
    def action_type(self):
        """
        :rtype: str
        """
        raise NotImplementedError

    @property
    def source_id(self):
        """
        :rtype: str
        """
        raise NotImplementedError

    @property
    def target_id(self):
        """
        :rtype: str
        """
        raise NotImplementedError

    @property
    def action_id(self):
        """
        :rtype: str
        """
        raise NotImplementedError

    @property
    def time_stamp(self):
        """
        :rtype: datetime
        """
        raise NotImplementedError

    @property
    def user_id(self):
        """
        :rtype: str
        """
        raise NotImplementedError

    @property
    def session_id(self):
        """
        :rtype: str
        """
        raise NotImplementedError

    @property
    def meta(self):
        """
        :rtype: dict
        """
        raise NotImplementedError

    def template(self):
        """
        :rtype: IActionTemplate
        """
        raise NotImplementedError



from math import cos, sin, radians
from types import FunctionType

import matplotlib.pyplot as plt
from matplotlib.axis import Axis
from matplotlib.axes import Axes
from typing import List, Union

from ux.interfaces.i_location import ILocation


def new_axes(width: int = 16, height: int = 9):
    """
    :rtype: Axes
    """
    _, ax = plt.subplots(figsize=(width, height))
    return ax


def set_axis_tick_label_rotation(ax: Axis, rotation: int):
    """
    Set the rotation of axis tick labels.

    :param ax: The axis whose tick label rotation to set.
    :param rotation: The rotation value to set.
    """
    if ax.get_majorticklabels():
        plt.setp(ax.get_majorticklabels(), rotation=rotation)
    if ax.get_minorticklabels():
        plt.setp(ax.get_minorticklabels(), rotation=rotation)


def get_hist_index(bin_values: List[int]):
    """
    Return index labels for histogram bin values.

    :rtype:  List[str]
    """
    return [
        '{}%-{}%'.format(bin_values[b], bin_values[b + 1])
        for b in range(len(bin_values) - 1)
    ]


def point_distance(point_1: tuple, point_2: tuple):
    """
    Calculate the distance from point_1 to point_2.

    :rtype: float
    """
    return (
        (point_1[0] - point_2[0]) ** 2 +
        (point_1[1] - point_2[1]) ** 2
    ) ** 0.5


def circle_edge(point_1: tuple, point_2: tuple, radius: float, rotation_angle: float = 0.0):
    """
    Calculate the closest point on the edge of a circle centered at point_1 with radius to point_2.

    :rtype: tuple
    """
    distance = point_distance(point_1, point_2)
    edge_point = (
        point_1[0] + radius * (point_2[0] - point_1[0]) / distance,
        point_1[1] + radius * (point_2[1] - point_1[1]) / distance
    )
    if rotation_angle != 0:
        rotation_angle = radians(rotation_angle)
        ox, oy = point_1
        px, py = edge_point
        qx = ox + cos(rotation_angle) * (px - ox) - sin(rotation_angle) * (py - oy)
        qy = oy + sin(rotation_angle) * (px - ox) + cos(rotation_angle) * (py - oy)
        edge_point = qx, qy

    return edge_point


def get_color(color: Union[None, str, FunctionType], state: ILocation, default: str):
    """
    Return a matplotlib color string depending on the type of `color`

    :param color: `string` to pass through, `None` to use default or `callable` to use a function.
    :param state: The location to use if `color` is a `callable`.
    :param default: Default color value if `None` was passed.
    :rtype: str
    """
    if color is None:
        return default
    elif type(color) is str:
        return color
    elif callable(color):
        return color(state)
    else:
        raise TypeError


def transform_axis_tick_labels(ax: Axis, transformation: FunctionType):
    """
    Transforms the labels of each label along the axis by a transformation function.

    :param ax: The axis whose tick labels to transform.
    :param transformation: The transformation function e.g. `lambda t: t.split('T')[0]`.
    """
    ax.figure.canvas.draw()  # make sure the figure has been drawn so the labels are available to be got
    labels = ax.get_ticklabels()
    for l in labels:
        new_label = transformation(l.get_text())
        l.set_text(new_label)
    ax.set_ticklabels(labels)



from matplotlib.axes import Axes
from numpy import array, stack
from pandas import Series

from ux.calcs.basic_calcs.task_success import binary_task_success_rate
from ux.plots.helpers import new_axes


def plot_percentage_task_errors(condition_result: dict, ax: Axes = None):
    """
    Plot the percentage of errors on completing a task, for one or more scenarios e.g. conditions.

    :param condition_result: Dictionary {condition_name => condition results list (0 = no errors & 1 = error)}.
    :param ax: Optional matplotlib axes.
    :rtype: Axes
    """
    # calculate error rate
    error_rates = {
        condition_name: binary_task_success_rate(condition_result[condition_name])
        for condition_name in condition_result.keys()
    }
    means = Series({
        condition_name: error_rates[condition_name][0]
        for condition_name in error_rates.keys()
    })
    confidence_intervals = array([
        error_rates[condition_name][1]
        for condition_name in means.index
    ])
    errors = confidence_intervals.copy()
    lower = array(means - errors[:, 0])
    upper = array(errors[:, 1] - means)
    errors = stack([lower, upper])
    ax = ax or new_axes()
    means *= 100
    errors *= 100
    means.plot(kind='bar', ax=ax, yerr=errors, capsize=5)
    ax.set_xlabel('Condition')
    ax.set_ylabel('% Login Attempts with Error(s)')
    return ax



from numpy import inf


class KPI(object):

    def __init__(self, name: str, numerator: int, denominator: int,
                 numer_config: dict = None, denom_config: dict = None):
        """
        Defines a KPI metric, typically expressed as a percentage of users who did something.

        :param numerator: The number of users who did something.
        :param denominator: The set of users who did or did not do the thing.
        :param numer_config: Optional dict[split_name, filter_name] of config params used to calculate the numerator.
        :param denom_config: Optional dict[split_name, filter_name] of config params used to calculate the denominator.
        """
        self.name = name
        self.numerator = numerator
        self.denominator = denominator
        self.numer_config = numer_config
        self.denom_config = denom_config

    @property
    def proportion(self):
        if self.denominator == 0:
            return inf
        return self.numerator / self.denominator

    @property
    def percentage(self):
        if self.denominator == 0:
            return inf
        return 100 * self.numerator / self.denominator

    def to_dict(self):
        """
        :rtype: Dict 
        """
        out_dict = {
            'name': self.name,
            'numerator': self.numerator,
            'denominator': self.denominator,
            'percentage': self.percentage,
            'proportion': self.proportion
        }
        if self.numer_config is not None:
            for k, v in self.numer_config.items():
                out_dict['numerator__' + k] = v
        if self.denom_config is not None:
            for k, v in self.denom_config.items():
                out_dict['denominator__' + k] = v
        return out_dict



from typing import List, Callable

from ux.calcs.object_calcs.efficiency import lostness
from ux.calcs.object_calcs.task_success import unordered_task_completion_rate, ordered_task_completion_rate, \
    binary_task_success
from ux.calcs.object_calcs.utils import sequence_intersects_task
from ux.interfaces.sequences.i_action_sequence import IActionSequence
from ux.interfaces.actions.i_action_template import IActionTemplate
from ux.interfaces.tasks.i_task import ITask


class Task(ITask):
    """
    Represents a Task corresponding to a sequence of desirable Actions.
    """
    def __init__(self, name: str, action_templates: List[IActionTemplate] = None):
        """
        Create a new Task.

        :param name: The name of the new Task.
        :param action_templates: A list of ActionTemplates that compose the Task.
        """
        self._name = name
        self._action_templates = action_templates or []

    @property
    def name(self):
        """
        Return the name of the Task.

        :rtype: str
        """
        return self._name

    @property
    def action_templates(self):
        """
        Return the list of ActionTemplates that compose the Task.

        :rtype: List[IActionTemplate]
        """
        return self._action_templates

    def add_action_template(self, action_template: IActionTemplate):
        """
        Add a new ActionTemplate to the end of the Task.

        :param action_template: The ActionTemplate to add.
        """
        self._action_templates.append(action_template)

    def unordered_completion_rate(self, action_sequence: IActionSequence):
        """
        Calculate the unordered completion rate of the Task from the Actions in the given ActionSequence.

        :rtype: float
        """
        return unordered_task_completion_rate(self, action_sequence)

    def ordered_completion_rate(self, action_sequence: IActionSequence):
        """
        Calculate the ordered completion rate of the Task from the Actions in the given ActionSequence.

        :rtype: float
        """
        return ordered_task_completion_rate(self, action_sequence)

    def intersects_sequence(self, action_sequence: IActionSequence):
        """
        Return True if the given ActionSequence has any ActionTemplates that are equivalent to any Actions in the Task.

        :rtype: bool
        """
        return sequence_intersects_task(action_sequence=action_sequence, task=self)

    def binary_task_success(self, action_sequence: IActionSequence,
                            success_func: Callable[[ITask, IActionSequence], bool]):
        """
        Return True if success_func is met.

        :param action_sequence: The ActionSequence to assess success of.
        :param success_func: Callable to use to assess success.
        :rtype: bool
        """
        return binary_task_success(
            task=self, action_sequence=action_sequence,
            success_func=success_func
        )

    def lostness(self, action_sequence: IActionSequence):
        """
        Return the lostness of the given ActionSequence with respect to this Task.

        :param action_sequence: The ActionSequence to calculate lostness for.
        :rtype: float
        """
        return lostness(task=self, action_sequence=action_sequence)

    def __len__(self):
        """
        :rtype: int
        """
        return len(self._action_templates)

    def __repr__(self):
        """
        :rtype: str
        """
        return 'Task({} [{}])'.format(
            self._name, len(self._action_templates)
        )

    def __getitem__(self, item):
        """
        :rtype: IUserAction
        """
        return self._action_templates[item]

    def __contains__(self, item):
        """
        :rtype: bool 
        """
        if isinstance(item, IActionTemplate):
            return item in self._action_templates
        else:
            raise TypeError('item must be IActionTemplate')

    def __iter__(self):
        """
        :rtype: IActionTemplate 
        """
        return self._action_templates.__iter__()



from collections import OrderedDict
import matplotlib.pyplot as plt
from numpy.random import random, normal
from numpy import add, min

from ux.plots.usability.efficiency import plot_task_completion_efficiency


def task_success_per_time_unit():
    """
    Emulate the plot from Figure 4.12 of "Measuring the User Experience"
    """
    # generate example data
    success_task_per_minute = OrderedDict(
        ('Prototype {}'.format(i),
         add(normal(random() * 2, 0.5, 100),
             abs(min(normal(random() * 2, 0.5, 100)))))
        for i in range(1, 5)
    )
    # plot
    plot_task_completion_efficiency(success_task_per_minute)
    plt.show()


if __name__ == '__main__':

    task_success_per_time_unit()



from typing import List

from ux.interfaces.i_user import IUser


class User(IUser):
    """
    Represents a User.
    """
    def __init__(self, user_id: str, session_ids: list = None, action_ids: list = None):
        """
        Create a new User.
        :param user_id: The User's id.
        :param session_ids: A list of id's of Sessions that the User took Actions in.
        :param action_ids: A list of the id's of Actions that the User took.
        """
        self._user_id = user_id
        self._session_ids = session_ids or []
        self._action_ids = action_ids or []

    @property
    def user_id(self):
        """
        Return the User's id

        :rtype: str
        """
        return self._user_id

    @property
    def action_ids(self):
        """
        Return the list of id's of the Actions the User took.

        :rtype: List[str]
        """
        return self._action_ids

    @property
    def session_ids(self):
        """
        Return the list of id's of the Sessions the User took Actions in.

        :rtype: List[str]
        """
        return self._session_ids

    def add_action_id(self, action_id: str):
        """
        Add an Action id to the list of the User's Actions.

        :param action_id: The id of the Action to add.
        """
        self._action_ids.append(action_id)

    def add_session_id(self, session_id: str):
        """
        Add a Session id to the list of the User's Sessions.

        :param session_id: The id of the Action to add.
        """
        self._session_ids.append(session_id)

    def __repr__(self):

        return 'User({})'.format(self._user_id)



from datetime import datetime

from ux.interfaces.actions.i_user_action import IUserAction
from ux.interfaces.actions.i_action_template import IActionTemplate
from ux.classes.actions.action_template import ActionTemplate


class UserAction(IUserAction):
    """
    Represents an Action taken by a User.
    """
    def __init__(self, action_id: str, action_type: str, source_id: str,
                 time_stamp: datetime, user_id: str, session_id: str,
                 target_id: str = None, meta: dict = None):
        """
        Create a new UserAction.

        :param action_id: The id of the Action taken.
        :param action_type: The type of Action taken.
        :param source_id: The id of the location where the User took the Action.
        :param time_stamp: The date-time when the User took the Action.
        :param user_id: The id of the User who took the Action.
        :param session_id: The id of the Session in which the User took the Action.
        :param target_id: Optional id of the location the Action navigated to.
        :param meta: Optional additional data to store with the UserAction.
        """
        self._action_type = action_type
        self._source_id = source_id
        self._target_id = target_id
        self._action_id = action_id
        self._time_stamp = time_stamp
        self._user_id = user_id
        self._session_id = session_id
        self._meta = meta
        self._action_template = None

    @property
    def action_type(self):
        """
        Return the type of Action taken.

        :rtype: str
        """
        return self._action_type

    @property
    def source_id(self):
        """
        Return the id of the location where the User took the Action.

        :rtype: str
        """
        return self._source_id

    @property
    def target_id(self):
        """
        Return the id of the location where the Action took the User to.

        :rtype: str
        """
        return self._target_id

    @property
    def action_id(self):
        """
        Return the id of the Action taken.

        :rtype: str
        """
        return self._action_id

    @property
    def time_stamp(self):
        """
        Return the date-time the Action was taken.

        :rtype: datetime
        """
        return self._time_stamp

    @property
    def user_id(self):
        """
        Return the id of the User who took the Action.

        :rtype: str
        """
        return self._user_id

    @property
    def session_id(self):
        """
        Return the id of the Session in which the User took the Action.

        :rtype: str
        """
        return self._session_id

    @property
    def meta(self):
        """
        Return any additional data stored with the UserAction.

        :rtype: dict
        """
        return self._meta

    def template(self):
        """
        Return an ActionTemplate that corresponds to the Action.

        :rtype: IActionTemplate
        """
        if self._action_template is None:
            self._action_template = ActionTemplate(
                action_type=self._action_type,
                source_id=self._source_id,
                target_id=self._target_id
            )
        return self._action_template

    def __repr__(self):

        return 'UserAction({}: {}{}{})'.format(
            self._action_type, self._source_id,
            ' -> ' if self._target_id else '',
            self._target_id if self._target_id else ''
        )



from collections import defaultdict
from typing import Dict, List

from ux.custom_types import SequenceFilter, ActionFilter, SequenceGrouper
from ux.interfaces.sequences.i_action_sequence import IActionSequence
from ux.classes.counts.count_config import CountConfig
from ux.classes.counts.temporal_count import TemporalCount


def count_actions_where(sequences: List[IActionSequence],
                        action_condition: ActionFilter,
                        sequence_condition: SequenceFilter = None,
                        split_by: SequenceGrouper = None):
    """
    Count the number of UserActions in the ActionSequences where the given condition is True.

    :param sequences: The list of IActionSequences to test.
    :param action_condition: Condition to evaluate each user against.
    :param sequence_condition: Optional condition to evaluate each sequence against before checking action_condition.
    :param split_by: Optional callable to split counts by some attribute of each action.
    """
    # allow all sequences through if no sequence condition is specified
    if sequence_condition is None:
        sequence_condition = lambda seq: True
    if split_by is None:
        count = 0
        for sequence in sequences:
            if sequence_condition(sequence):
                for action in sequence:
                    if action_condition(action):
                        count += 1
        return count
    else:
        counts = defaultdict(int)
        for sequence in sequences:
            if sequence_condition(sequence):
                for action in sequence.user_actions:
                    if action_condition(action):
                        split_result = split_by(action)
                        if type(split_result) is list:
                            for result in split_result:
                                counts[result] += 1
                        else:
                            counts[split_result] += 1
        return dict(counts)


def count_sequences_where(sequences: List[IActionSequence], condition, split_by: SequenceGrouper = None):
    """
    Count the number of ActionSequences in the list where the given condition is True.

    :param sequences: The list of IActionSequences to test.
    :param condition: The condition to evaluate each sequence against.
    :type condition: SequenceFilter
    :param split_by: Optional callable to split counts by some attribute of each sequence. Should return a string.
    :return: Integer count if split_by is None. Otherwise dict of {split_value: count}
    """
    if split_by is None:
        count = 0
        for sequence in sequences:
            if condition(sequence):
                count += 1
        return count
    else:
        counts = defaultdict(int)
        for sequence in sequences:
            if condition(sequence):
                counts[split_by(sequence)] += 1
        return dict(counts)


def get_sequences_where(sequences: List[IActionSequence], condition, split_by: SequenceGrouper = None):
    """
    Get the ActionSequences in the list where the given condition is True.

    :param sequences: The list of IActionSequences to test.
    :param condition: The condition to evaluate each sequence against.
    :type condition: SequenceFilter
    :param split_by: Optional callable to split counts by some attribute of each sequence. Should return a string.
    :return: List of sequences if split_by is None. Otherwise dict of {split_value: sequences}
    """
    if split_by is None:
        filtered = []
        for sequence in sequences:
            if condition(sequence):
                filtered.append(sequence)
        return filtered
    else:
        filtered = defaultdict(list)
        for sequence in sequences:
            if condition(sequence):
                filtered[split_by(sequence)].append(sequence)
        return dict(filtered)


def temporal_counts_by_config(sequences: List[IActionSequence], configs: List[CountConfig],
                              temporal_split):
    """
    Count metrics using the settings in a list of CountConfigs.

    :param sequences: List of ActionSequences containing actions to measure metrics.
    :param configs: List of CountConfigs defining the metrics to count.
    :param temporal_split: lambda function returning OrderedDict[date, List[IActionSequence]]
    :type temporal_split: Callable[[List[IActionSequence]], List[IActionSequence]]
    :rtype: Dict[str, TemporalCount]
    :return Dict[config.name, TemporalCount for config]
    """
    total_counts = dict()
    for config in configs:
        total_counts[config.name] = TemporalCount(config.name)
    sequence_groups = temporal_split(sequences)
    for sequence_date, date_sequences in sequence_groups.items():
        for config in configs:
            # count sequences
            if config.action_condition is None:
                counts = count_sequences_where(
                    sequences=date_sequences,
                    condition=config.sequence_condition,
                    split_by=config.sequence_split_by
                )
            # count actions
            else:
                counts = count_actions_where(
                    sequences=date_sequences,
                    sequence_condition=config.sequence_condition,
                    action_condition=config.action_condition,
                    split_by=config.action_split_by
                )
            total_counts[config.name][sequence_date] = counts

    return total_counts



from math import sqrt

from ux.interfaces.sequences.i_action_sequence import IActionSequence
from ux.interfaces.tasks.i_task import ITask


def lostness(task: ITask, action_sequence: IActionSequence):
    """
    Calculate the `lostness` metric for the given ActionSequence, using the Task as a reference.

    :param task: The Task to use to find the optimum number of Actions.
    :param action_sequence: The ActionSequence to use to find the unique and total number of Actions.
    :rtype: float
    """
    task_templates = task.action_templates
    assert len(task_templates) == len(set(task_templates)), \
        'Task contains repeated ActionTemplates'
    optimum = len(task.action_templates)
    sequence_templates = action_sequence.action_templates()
    unique = len(set(sequence_templates))
    total = len(sequence_templates)
    return sqrt(
        (unique / total - 1) ** 2 +
        (optimum / unique - 1) ** 2
    )



from collections import defaultdict, OrderedDict
from pandas import DataFrame, Series
from types import FunctionType
from typing import Dict, List, Union

from ux.classes.wrappers.map_result import MapResult
from ux.custom_types import SequenceFilter, SequencesGroupByKey, SequenceFilterSet, SequencesGrouper, SequenceGrouper
from ux.interfaces.sequences.i_sequences import ISequences
from ux.interfaces.sequences.i_sequences_group_by import ISequencesGroupBy
from ux.utils.misc import get_method_name


class SequencesGroupBy(ISequencesGroupBy):

    def __init__(self, data: Dict[SequencesGroupByKey, ISequences], names: Union[str, List[str]]):
        """
        :param data: Dictionary mapping keys to Sequences collections
        :param names: Names for the key groups
        """
        self._data: Dict[SequencesGroupByKey, ISequences] = OrderedDict(data)
        if type(names) is str:
            names = [names]
        self._names: List[str] = names

        for attribute in self._data.keys():
            try:
                setattr(self, attribute, self._data[attribute])
            except:
                pass

    def count(self):
        """
        :rtype: MapResult
        """
        out_dict = OrderedDict([
            (key, sequences.count())
            for key, sequences in self._data.items()
        ])
        return MapResult(out_dict, key_names=self.names, value_names='count')

    def map(self, mapper: Union[str, dict, list, SequencesGrouper]):
        """
        Apply a map function to every Sequences in the GroupBy and return the results.

        :param mapper: The method or methods to apply to each Sequences
        :rtype: MapResult
        """
        def map_items(item_mapper: Union[str, FunctionType]):
            if isinstance(item_mapper, str):
                # properties and methods
                if hasattr(ISequences, item_mapper):
                    if callable(getattr(ISequences, item_mapper)):
                        # methods
                        return [getattr(sequences, item_mapper)() for sequences in self._data.values()]
                    else:
                        # properties
                        return [getattr(sequences, item_mapper) for sequences in self._data.values()]
            elif isinstance(item_mapper, FunctionType):
                return [item_mapper(sequences) for sequences in self._data.values()]
            else:
                raise TypeError('item mappers must be FunctionType or str')

        group_names = list(self._data.keys())

        def new_group(names, method):
            if isinstance(names, str):
                names = [names]
            return tuple(names + [get_method_name(method)])

        if isinstance(mapper, str) or isinstance(mapper, FunctionType):
            keys = [new_group(names, mapper) for names in group_names]
            values = map_items(mapper)
            results = OrderedDict([(key, value) for key, value in zip(keys, values)])
        elif isinstance(mapper, dict):
            results = OrderedDict()
            for map_key, map_func in mapper.items():
                keys = [new_group(names, map_key) for names in group_names]
                values = map_items(map_func)
                results.update(OrderedDict([
                    (key, value) for key, value in zip(keys, values)
                ]))
        elif isinstance(mapper, list):
            results = OrderedDict()
            for map_item in mapper:
                keys = [new_group(names, map_item) for names in group_names]
                values = map_items(map_item)
                results.update(OrderedDict([
                    (key, value) for key, value in zip(keys, values)
                ]))
        else:
            raise TypeError('mapper must be dict, list, str or FunctionType')

        return MapResult(results, key_names=self.names + ['map'])

    def agg(self, agg_funcs: dict):
        """
        :param agg_funcs: dict mapping attributes to one or more aggregation functions e.g. duration -> np.median
        """
        results = defaultdict(dict)
        # build list of agg_key, agg_func pairs
        agg_pairs = []
        for k, v in agg_funcs.items():
            if isinstance(v, FunctionType):
                agg_pairs.append((k, v))
            else:
                for f in v:
                    agg_pairs.append((k, f))
        for agg_key, agg_func in agg_pairs:
            agg_name = get_method_name(agg_func)
            for name, sequences in self.items():
                if hasattr(ISequences, agg_key):
                    if callable(getattr(ISequences, agg_key)):
                        values = getattr(sequences, agg_key)()
                    else:
                        values = getattr(sequences, agg_key)
                    results[name]['{}({})'.format(agg_name, agg_key)] = agg_func(values)
        return results

    def filter(self, condition: SequenceFilter):
        """
        Return a new Sequences containing only the sequences matching the `condition` in each group.

        :param condition: lambda(sequence) that returns True to include a sequence.
        :rtype: ISequencesGroupBy
        """
        data = {
            key: value.filter(condition)
            for key, value in self._data.items()
        }
        return SequencesGroupBy(data=data, names=self.names)

    def group_filter(self, filters: SequenceFilterSet, group_name: str = None):
        """
        Return a new SequencesGroupBy keyed by the filter name with values matching each filter, applied in parallel.

        :param filters: Dictionary of filters to apply.
        :param group_name: Name to identify the filter group.
        :rtype: SequencesGroupBy
        """
        names = self._names.copy()
        if group_name is None:
            i = 2
            while 'filter_' + str(i) in names:
                i += 1
            group_name = 'filter_' + str(i)
        if group_name in names:
            raise ValueError(f'Name "{group_name}" already exists in SequencesGroupBy instance.')
        else:
            names.append(group_name)
        data = {}
        for data_filter_names, data_sequences in self._data.items():
            for filter_name, filter_condition in filters.items():
                if type(data_filter_names) is str:
                    new_filter_names = (data_filter_names, filter_name)
                else:
                    new_filter_names = tuple(list(data_filter_names) + [filter_name])
                data[new_filter_names] = data_sequences.filter(filter_condition)
        return SequencesGroupBy(data=data, names=names)

    def group_by(self, by: Union[SequenceGrouper, Dict[str, SequenceGrouper], str, list]):
        """
        Return a new SequencesGroupBy keyed by each value returned by a single grouper,
        or each combination of groupers for a list of groupers.
        Each grouper should be a lambda function that returns a picklable value e.g. str.

        :param by: lambda(Sequence) or dict[group_name, lambda(Sequence)] or list[str or lambda(Sequence)].
        :rtype: SequencesGroupBy
        """
        new_results = OrderedDict()
        for group_key, group_sequences in self.items():
            if isinstance(group_key, tuple):
                group_key = list(group_key)
            else:
                group_key = [group_key]
            group_sub_results: ISequencesGroupBy = group_sequences.group_by(by)
            for group_sub_key, group_sub_values in group_sub_results.items():
                if isinstance(group_sub_key, tuple):
                    group_sub_key = list(group_sub_key)
                else:
                    group_sub_key = [group_sub_key]
                new_results[tuple(group_key + group_sub_key)] = group_sub_values

        return SequencesGroupBy(data=new_results, names=self.names + group_sub_results.names)

    def items(self):
        """
        :rtype: dict_items
        """
        return self._data.items()

    def keys(self):
        """
        :rtype: dict_keys
        """
        return self._data.keys()

    def values(self):
        """
        :rtype: dict_values
        """
        return self._data.values()

    @property
    def names(self):
        """
        :rtype: List[str]
        """
        return self._names
    
    @names.setter
    def names(self, names):

        self._names = names

    def __getitem__(self, item):
        """
        :rtype: ISequences
        """
        return self._data[item]

    def __repr__(self):

        return 'SequencesGroupBy({})'.format(self._data.__repr__())

    def __len__(self):

        return len(self._data)

    def __contains__(self, item):

        return item in self._data

    def __iter__(self):

        return self._data.__iter__()



from collections import OrderedDict
import matplotlib.pyplot as plt
from numpy.random import randint, normal
from pandas import DataFrame

from ux.plots.usability.time_on_task import plot_task_completion_times, plot_task_completion_under_threshold


def mean_time_on_task():
    """
    Emulate the plot from Figure 4.5 of "Measuring the User Experience".
    """
    # generate example data
    task_time = OrderedDict(
        ('Task {}'.format(i), normal(randint(60, 100), 10, 100))
        for i in range(1, 11)
    )
    # plot
    plot_task_completion_times(task_time)
    plt.show()


def task_completion_under_threshold():
    """
    Emulate the plot from Figure 4.6 of "Measuring the User Experience".
    """
    # generate example data
    participant_times = {
       'P1': [259, 112, 135, 58, 8],
       'P2': [253, 64, 278, 160, 22],
       'P3': [42, 51, 60, 57, 26],
       'P4': [38, 108, 115, 146, 26],
       'P5': [33, 142, 66, 47, 38],
       'P6': [33, 54, 261, 26, 42],
       'P7': [36, 152, 53, 22, 44],
       'P8': [112, 65, 171, 133, 46],
       'P9': [29, 92, 147, 56, 56],
       'P10': [158, 113, 136, 83, 64],
       'P11': [24, 69, 119, 25, 68],
       'P12': [108, 50, 145, 15, 75],
       'P13': [110, 128, 97, 97, 78],
       'P14': [37, 66, 105, 83, 80],
       'P15': [116, 78, 40, 163, 100],
       'P16': [129, 152, 67, 168, 109],
       'P17': [31, 51, 51, 119, 116],
       'P18': [33, 97, 44, 81, 127],
       'P19': [75, 124, 286, 103, 236],
       'P20': [76, 62, 108, 185, 245]
    }
    df = DataFrame.from_dict(participant_times, orient='index')
    df.columns = ['Task 1', 'Task 2', 'Task 3', 'Task 4', 'Task 5']
    participant_times = df.to_dict()
    # plot
    plot_task_completion_under_threshold(participant_times)
    plt.show()


if __name__ == '__main__':

    mean_time_on_task()
    task_completion_under_threshold()




from unittest import TestCase

from pandas import Series, Index, DataFrame

from ux.classes.wrappers.map_result import MapResult


class TestMapResult(TestCase):

    def setUp(self) -> None:

        self.mr_single_single = MapResult(
            data={'a': 1, 'b': 2, 'c': 3},
            key_names='letters',
            value_names='numbers'
        )
        self.s_single_single = Series(
            index=Index(data=['a', 'b', 'c'], name='letters'),
            data=[1, 2, 3], name='numbers'
        )
        self.mr_single_fixed = MapResult(
            data={'a': [1, 2, 3], 'b': [4, 5, 6]},
            key_names='letters',
            value_names='numbers'
        )
        self.s_single_fixed = Series(
            index=Index(data=['a', 'a', 'a', 'b', 'b', 'b'], name='letters'),
            data=[1, 2, 3, 4, 5, 6], name='numbers'
        )
        self.d_data_fixed_wide = DataFrame(
            data={'a': [1, 2, 3], 'b': [4, 5, 6]}
        )
        self.mr_single_variable = MapResult(
            data={'a': [1, 2], 'b': [3, 4, 5]},
            key_names='letters',
            value_names='numbers'
        )
        self.s_single_variable = Series(
            index=Index(data=['a', 'a', 'b', 'b', 'b'], name='letters'),
            data=[1, 2, 3, 4, 5], name='numbers'
        )
        self.mr_tuple_single = MapResult(
            data={('a', 'b'): 1, ('c', 'd'): 2, ('e', 'f'): 3},
            key_names=['letter_1', 'letter_2'],
            value_names='numbers'
        )
        self.s_tuple_single = Series(
            index=Index(data=[('a', 'b'), ('c', 'd'), ('e', 'f')], names=['letter_1', 'letter_2']),
            data=[1, 2, 3], name='numbers'
        )
        self.mr_tuple_fixed = MapResult(
            data={('a', 'b'): [1, 2, 3], ('c', 'd'): [4, 5, 6]},
            key_names=['letter_1', 'letter_2'],
            value_names='numbers'
        )
        self.s_tuple_fixed = Series(
            index=Index(data=[('a', 'b'), ('a', 'b'), ('a', 'b'), ('c', 'd'), ('c', 'd'), ('c', 'd')],
                        names=['letter_1', 'letter_2']),
            data=[1, 2, 3, 4, 5, 6], name='numbers'
        )
        self.mr_tuple_variable = MapResult(
            data={('a', 'b'): [1, 2], ('c', 'd'): [3, 4, 5]},
            key_names=['letter_1', 'letter_2'],
            value_names='numbers'
        )
        self.s_tuple_variable = Series(
            index=Index(data=[('a', 'b'), ('a', 'b'), ('c', 'd'), ('c', 'd'), ('c', 'd')],
                        names=['letter_1', 'letter_2']),
            data=[1, 2, 3, 4, 5], name='numbers'
        )

    @staticmethod
    def series_equivalent(data_1: Series, data_2: Series):

        return (
            data_1.index.tolist() == data_2.index.tolist() and
            data_1.to_list() == data_2.to_list()
        )

    @staticmethod
    def frames_equivalent(data_1: DataFrame, data_2: DataFrame):

        return (
            sorted(data_1.columns) == sorted(data_2.columns) and
            data_1.index.to_list() == data_2.index.to_list()
            and all(data_1[column].to_list() == data_2[column].to_list() for column in data_1.columns)
        )

    def test_to_series(self):

        self.assertTrue(self.series_equivalent(self.s_single_single, self.mr_single_single.to_series()))
        self.assertTrue(self.series_equivalent(self.s_single_fixed, self.mr_single_fixed.to_series()))
        self.assertTrue(self.series_equivalent(self.s_single_variable, self.mr_single_variable.to_series()))
        self.assertTrue(self.series_equivalent(self.s_tuple_single, self.mr_tuple_single.to_series()))
        self.assertTrue(self.series_equivalent(self.s_tuple_fixed, self.mr_tuple_fixed.to_series()))
        self.assertTrue(self.series_equivalent(self.s_tuple_variable, self.mr_tuple_variable.to_series()))

    def test_to_frame(self):

        self.assertTrue(self.frames_equivalent(self.s_single_single.reset_index(), self.mr_single_single.to_frame()))
        self.assertTrue(self.frames_equivalent(self.s_single_fixed.reset_index(), self.mr_single_fixed.to_frame()))
        self.assertTrue(self.frames_equivalent(self.s_single_variable.reset_index(), self.mr_single_variable.to_frame()))
        self.assertTrue(self.frames_equivalent(self.s_tuple_single.reset_index(), self.mr_tuple_single.to_frame()))
        self.assertTrue(self.frames_equivalent(self.s_tuple_fixed.reset_index(), self.mr_tuple_fixed.to_frame()))
        self.assertTrue(self.frames_equivalent(self.s_tuple_variable.reset_index(), self.mr_tuple_variable.to_frame()))

    def test_to_frame_wide(self):

        self.assertTrue(self.frames_equivalent(self.d_data_fixed_wide, self.mr_single_fixed.to_frame(wide=True)))



from ux.interfaces.sequences.i_action_sequence import IActionSequence
from ux.interfaces.tasks.i_task import ITask


def sequence_intersects_task(action_sequence: IActionSequence, task: ITask):
    """
    Determine if the actions in the sequence intersect with those in the task.

    :rtype: bool
    """
    sequence_template_set = set(action_sequence.action_templates())
    task_template_set = set(task.action_templates)
    return len(sequence_template_set.intersection(task_template_set)) > 0



from collections import defaultdict, OrderedDict, Counter
from datetime import timedelta
from itertools import chain, product
from pandas import notnull
from types import FunctionType
from typing import Dict, Iterator, List, Union, Tuple

from ux.classes.sequences.sequences_group_by import SequencesGroupBy
from ux.classes.wrappers.map_result import MapResult
from ux.custom_types import SequenceFilter, SequenceFilterSet, SequenceGrouper
from ux.interfaces.sequences.i_action_sequence import IActionSequence
from ux.interfaces.sequences.i_sequences import ISequences
from ux.utils.misc import get_method_name


class Sequences(ISequences):

    _sequence_lookups = {
        'date': lambda seq: seq.start.date(),
        'start_date': lambda seq: seq.start.date(),
        'end_date': lambda seq: seq.end.date(),
        'date_time': lambda seq: seq.start,
        'start': lambda seq: seq.start,
        'end': lambda seq: seq.end,
        'hour': lambda seq: seq.start.hour,
        'start_hour': lambda seq: seq.start.hour,
        'end_hour': lambda seq: seq.end.hour,
        'day': lambda seq: seq.start.day,
        'start_day': lambda seq: seq.start.day,
        'end_day': lambda seq: seq.end.day,
        'weekday': lambda seq: seq.start.isoweekday(),
        'start_weekday': lambda seq: seq.start.isoweekday(),
        'end_weekday': lambda seq: seq.end.isoweekday(),
        'week': lambda seq: seq.start.isocalendar()[1],
        'start_week': lambda seq: seq.start.isocalendar()[1],
        'end_week': lambda seq: seq.end.isocalendar()[1],
        'month': lambda seq: seq.start.month,
        'start_month': lambda seq: seq.start.month,
        'end_month': lambda seq: seq.end.month
    }

    def __init__(self, sequences: List[IActionSequence]):
        """
        Create a new Sequences collection.

        :param sequences: List of ActionSequences to use to create the object.
        """
        self._sequences: List[IActionSequence] = sequences

    @property
    def sequences(self):
        """
        Return a list of the individual Sequences in the collection.

        :rtype: List[IActionSequence]
        """
        return self._sequences

    def filter(self, condition: SequenceFilter):
        """
        Return a new Sequences containing only the sequences matching the `condition`.

        :param condition: lambda(sequence) that returns True to include a sequence.
        :rtype: ISequences
        """
        filtered = []
        for sequence in self:
            if condition(sequence):
                filtered.append(sequence)
        return Sequences(filtered)

    def group_filter(self, filters: SequenceFilterSet, group_name: str = 'filter'):
        """
        Return a new SequencesGroupBy keyed by the filter name with values matching each filter, applied in parallel.

        :param filters: Dictionary of filters to apply.
        :param group_name: Name to identify the filter group.
        :rtype: SequencesGroupBy
        """
        filtered = {
            filter_name: self.filter(filter_condition)
            for filter_name, filter_condition in filters.items()
        }
        return SequencesGroupBy(data=filtered, names=[group_name])

    def chain_filter(self, filters: SequenceFilterSet):
        """
        Return a new SequencesGroupBy keyed by the dict key with values matching each filter, applied in series.

        :param filters: Dictionary of filters to apply. Use OrderedDict for Python < 3.7 to preserve key order.
        :rtype: SequencesGroupBy
        """
        filtered = OrderedDict()
        sequences = self._sequences
        for filter_name, filter_func in filters.items():
            filtered_sequences = []
            for sequence in sequences:
                if filter_func(sequence):
                    filtered_sequences.append(sequence)
            filtered[filter_name] = Sequences(filtered_sequences)
            sequences = filtered_sequences
        return SequencesGroupBy(data=filtered, names=['filter'])

    def group_by(self, by: Union[SequenceGrouper, Dict[str, SequenceGrouper], str, list]):
        """
        Return a SequencesGroupBy keyed by each value returned by a single grouper,
        or each combination of groupers for a list of groupers.
        Each grouper should be a lambda function that returns a picklable value e.g. str.

        :param by: lambda(Sequence) or dict[group_name, lambda(Sequence)] or list[str or lambda(Sequence)].
        :rtype: SequencesGroupBy
        """
        def apply_group_by(method: SequenceGrouper):
            """
            :rtype: Dict[str, Sequences]
            """
            splits = defaultdict(list)
            for sequence in self:
                splits[method(sequence)].append(sequence)
            return {
                group_name: Sequences(group_sequences)
                for group_name, group_sequences in splits.items()
            }

        # build groupers dict mapping name to grouping function
        groupers = OrderedDict()
        if isinstance(by, FunctionType):
            groupers[by.__name__] = by
        elif isinstance(by, str):
            if by not in self._sequence_lookups.keys():
                raise ValueError('Cannot group by "{}"'.format(by))
            else:
                groupers[by] = self._sequence_lookups[by]
        elif isinstance(by, dict):
            groupers = by
        elif isinstance(by, list):
            for element in by:
                if isinstance(element, str):
                    if element not in self._sequence_lookups.keys():
                        raise ValueError('Cannot group by "{}"'.format(element))
                    else:
                        groupers[element] = self._sequence_lookups[element]
                elif isinstance(element, FunctionType):
                    method_name = get_method_name(element)
                    groupers[method_name] = element
                else:
                    raise TypeError('List elements must be strings or functions.')

        # calculate dict mapping group-by split names to Sequences
        group_by_names = list(groupers.keys())
        group_bys = OrderedDict([
            (by_name, apply_group_by(groupers[by_name]))
            for by_name in group_by_names
        ])
        result = dict()
        for subgroups_combo in product(*[group_bys[group].items() for group in group_bys.keys()]):
            result_key = tuple([subgroup[0] for subgroup in subgroups_combo])
            if len(result_key) == 1:
                result_key = result_key[0]
            result_sequences = [subgroup[1] for subgroup in subgroups_combo]
            result[result_key] = Sequences.intersect_all(result_sequences)
        return SequencesGroupBy(result, names=group_by_names)

    def map(self, mapper: Union[str, dict, list, SequenceGrouper]):
        """
        Apply a map function to every Sequence in the Sequences and return the results.

        :param mapper: The method or methods to apply to each ActionSequence
        :rtype: MapResult
        """
        def map_items(item_mapper: Union[str, FunctionType]):
            if isinstance(item_mapper, str):
                # properties and methods
                if hasattr(IActionSequence, item_mapper):
                    if callable(getattr(self[0], item_mapper)):
                        # methods
                        return [getattr(sequence, item_mapper)() for sequence in self]
                    else:
                        # properties
                        return [getattr(sequence, item_mapper) for sequence in self]
                elif item_mapper in self._sequence_lookups:
                    return [self._sequence_lookups[item_mapper](sequence) for sequence in self]
                else:
                    raise ValueError(f'IActionSequence has no property or attribute named {item_mapper}')
            elif isinstance(item_mapper, FunctionType):
                return [item_mapper(sequence) for sequence in self]
            else:
                raise TypeError('item mappers must be FunctionType or str')

        if isinstance(mapper, str) or isinstance(mapper, FunctionType):
            results = OrderedDict([(get_method_name(mapper), map_items(mapper))])
        elif isinstance(mapper, dict):
            results = OrderedDict([
                (get_method_name(key), map_items(value))
                for key, value in mapper.items()
            ])
        elif isinstance(mapper, list):
            results = OrderedDict([
                (get_method_name(item), map_items(item))
                for item in mapper
            ])
        else:
            raise TypeError('mapper must be dict, list, str or FunctionType')

        return MapResult(results)

    def count(self):
        """
        Return the number of ActionSequences in the collection.

        :rtype: int
        """
        return len(self._sequences)

    def copy(self):
        """
        Return a new collection referencing this collection's ActionSequences.

        :rtype: ISequences
        """
        return Sequences(self._sequences)

    def intersection(self, other):
        """
        Return a new collection representing the ActionSequences in both collections.

        :type other: Union[Sequences, List[IActionSequence]]
        :rtype: ISequences
        """
        if type(other) is Sequences:
            other = other.sequences
        return Sequences(
            list(set(self._sequences).intersection(other))
        )

    @staticmethod
    def intersect_all(sequences):
        """
        Return a new collection representing the ActionSequences in every collection.

        :type sequences: List[ISequences]
        :rtype: ISequences
        """
        intersect = sequences[0].copy()
        for s in sequences[1:]:
            intersect = intersect.intersection(s)
        return intersect

    def back_click_rates(self):
        """
        :rtype: Dict[IActionTemplate, float]
        """
        rates = {}
        counts = self.action_template_counts()
        for template, count in counts.items():
            forwards = count
            backwards = counts.get(template.reversed(), 0)
            rate = backwards / forwards
            if rate <= 1:
                rates[template] = rate
        return rates

    @property
    def durations(self):
        """
        :rtype: List[timedelta]
        """
        return [sequence.duration for sequence in self]

    def action_template_counts(self):
        """
        Return a total count of all the ActionTemplates in the ActionSequences in the collection.

        :rtype: Dict[IActionTemplate, int]
        """
        counts = defaultdict(int)
        for sequence in self:
            for template in sequence.action_templates():
                counts[template] += 1
        return dict(counts)

    def action_template_sequence_counts(self):
        """
        Return a total count of the number of ActionSequences containing each ActionTemplate in the collection.

        :rtype: Dict[IActionTemplate, int]
        """
        counts = Counter(chain.from_iterable(
            list(sequence.action_template_set()) for sequence in self
        ))
        return dict(counts)

    def action_template_transition_counts(self):
        """
        Return counts of transitions between pairs of Actions from each Sequence in the collection.

        :return: Dictionary of {(from, to) => count}
        :rtype: Dict[tuple[IActionTemplate, IActionTemplate], int]
        """
        transitions = defaultdict(int)
        # count transitions
        for sequence in self:
            for a in range(len(sequence) - 1):
                from_action = sequence.user_actions[a].template()
                to_action = sequence.user_actions[a + 1].template()
                transitions[(from_action, to_action)] += 1
        return dict(transitions)

    def location_transition_counts(self, exclude: Union[str, List[str]] = None):
        """
        Count the transitions from each location to each other location in actions in the given sequences.

        :return: Counter[Tuple[from, to], count]
        :rtype: Counter[Tuple[str, str], int]
        """
        transitions = Counter()
        if exclude is not None:
            if isinstance(exclude, str):
                exclude = [exclude]
        else:
            exclude = []
        # count transitions
        for sequence in self:
            for action in sequence:
                source = action.source_id
                target = action.target_id
                if notnull(target) and source not in exclude and target not in exclude:
                    transitions[(source, target)] += 1
        return transitions

    def dwell_times(self, sum_by_location: bool, sum_by_sequence: bool):
        """
        Return the amount of time spent by the user at each location.

        :param sum_by_location: Whether to sum the durations of time spent at each location or keep as a list.
        :param sum_by_sequence: Whether to sum the durations of time spent at each location in each sequence or keep as a list.
        :rtype: Dict[str, Union[timedelta, List[timedelta]]]
        """
        if sum_by_sequence:
            dwell_times = defaultdict(timedelta)
            for sequence in self:
                for location, duration in sequence.dwell_times(sum_by_location=True).items():
                    dwell_times[location] += duration
        elif not sum_by_location and not sum_by_sequence:
            dwell_times = defaultdict(list)
            for sequence in self:
                for location, durations in sequence.dwell_times(sum_by_location=False).items():
                    dwell_times[location].extend(durations)
        elif sum_by_location and not sum_by_sequence:
            dwell_times = defaultdict(list)
            for sequence in self:
                for location, duration in sequence.dwell_times(sum_by_location=True).items():
                    dwell_times[location].append(duration)
        else:
            raise ValueError('Invalid arguments for sum_per_location and / or sum_per_sequence')

        return dwell_times

    def most_probable_location_sequence(self, exclude: Union[str, List[str]] = None, start_at: str = None,
                                        allow_repeats: bool = True):
        """
        Find the most probable location sequence.

        :param exclude: Optional list of names to exclude from the sequence.
        :param start_at: Optional name of start state. Leave as None to use most common state.
        :param allow_repeats: Whether to stop building the sequence at the point where a previous item is found or to
                              use a less likely item instead and carry on building the sequence.
        :rtype: List[str]
        """
        transitions = self.location_transition_counts(exclude=exclude)
        # find most frequent from point
        if start_at is not None:
            current_name = start_at
        else:
            current_name = transitions.most_common(1)[0][0][0]
        sequence = [current_name]
        # iteratively find most frequent transition point from the current point
        found = True
        while found:
            # remove last added location as a possible target
            if not allow_repeats:
                transitions = {from_to: count for from_to, count in transitions.items()
                               if from_to[1] != current_name}
            # find next location
            froms = [from_to[0] for from_to in transitions.keys()]
            if current_name in froms:
                current_name = Counter({
                    from_to: count for from_to, count in transitions.items()
                    if from_to[0] == current_name
                }).most_common(1)[0][0][1]
                # add if not already in sequence
                if current_name not in sequence:
                    sequence.append(current_name)
                else:
                    found = False
            else:
                found = False
        return sequence

    def sort(self, by: str, ascending: bool = True):
        """
        :rtype: ISequences
        """
        return Sequences(sequences=sorted(
            self._sequences, key=self._sequence_lookups[by],
            reverse=not ascending
        ))

    def __getitem__(self, item):
        """
        :rtype: IActionSequence
        """
        return self._sequences[item]

    def __repr__(self):

        return 'Sequences({})'.format(len(self._sequences))

    def __len__(self):

        return len(self._sequences)

    def __contains__(self, item: IActionSequence):

        if isinstance(item, IActionSequence):
            return item in self
        else:
            raise TypeError('item must be IActionSequence')

    def __iter__(self):
        """
        :rtype: Iterator[IActionSequence]
        """
        return self._sequences.__iter__()

    def __add__(self, other):
        """
        :rtype: ISequences
        """
        if type(other) is Sequences:
            other = other.sequences
        return Sequences(
            list(set(self._sequences).union(other))
        ).sort('date_time')

    def __sub__(self, other):
        """
        :rtype: ISequences
        """
        if type(other) is Sequences:
            other = other.sequences
        return Sequences(list(
            set(self._sequences) - set(self._sequences).intersection(other)
        )).sort('date_time')



from matplotlib.axes import Axes
from numpy import array, mean
from pandas import Series

from ux.calcs.basic_calcs.stats import normal_confidence_interval
from ux.plots.helpers import new_axes


def plot_task_completion_efficiency(success_task_per_minute: dict, confidence: float = 0.95, ax: Axes = None):
    """
    Plot the number of completed tasks per unit time for each scenario e.g. prototypes, designs.

    :param success_task_per_minute: Dictionary of {scenario name => list of number of successful tasks per minute}
    :param confidence: The confidence interval to use for error bars (0 - 1)
    :param ax: Optional matplotlib axes to plot on.
    :rtype: Axes
    """
    means = Series({
        prototype_name: mean(num_success)
        for prototype_name, num_success in success_task_per_minute.items()
    })
    errors = array([
        normal_confidence_interval(success_task_per_minute[prototype_name], confidence=confidence)
        for prototype_name in means.index
    ]).reshape(1, 2, means.shape[0])
    ax = ax or new_axes()
    means.plot(kind='bar', ax=ax, yerr=errors, capsize=5)
    ax.set_xlabel('Prototypes')
    ax.set_ylabel('Tasks Successfully Completed per Minute')

    return ax




from pandas import DataFrame, notnull


def add_meta(template: dict, key: str, value: str):

    if 'meta' not in template.keys():
        template['meta'] = {}
    template['meta'][key] = value
    return template


def generate_generic_template(specifics: dict, placeholders: dict, meta: dict = None):
    """
    Generate an example template from a dictionary of keys and values.

    :param specifics: Keys and values specific to this template.
    :param placeholders: Keys and values for all templates.
    :param meta: Additional values for meta, other than those already in `specifics`.
    :rtype: dict
    """
    template = placeholders.copy()
    for key, value in specifics.items():
        if notnull(value):
            if key in ('source', 'destination'):
                template[key] = value
            elif key == 'meta':
                if 'meta' not in template.keys():
                    template['meta'] = {}
                for line in value.split('\n'):
                    template['meta'][line.split(':')[0]] = line.split(':')[1]
                if meta is not None:
                    for k_meta, v_meta in meta.items():
                        template['meta'][k_meta] = v_meta
    return template


def generate_checkbox_check_template(specifics: dict, placeholders: dict,
                                     **kwargs):
    """
    Generate a template for the check event of a checkbox.

    :rtype: dict
    """
    template = generate_generic_template(
        specifics=specifics, placeholders=placeholders,
        meta={'value': '{{checkbox-value}}'}
    )
    template['action-type'] = 'check'
    return template


def generate_checkbox_uncheck_template(specifics: dict, placeholders: dict,
                                       **kwargs):
    """
    Generate a template for the uncheck event of a checkbox.

    :rtype: dict
    """
    template = generate_generic_template(
        specifics=specifics, placeholders=placeholders,
        meta={'value': '{{checkbox-value}}'}
    )
    template['action-type'] = 'uncheck'
    return template


def generate_dropdown_select_template(specifics: dict, placeholders: dict,
                                      **kwargs):
    """
    Generate a template for the select event of a dropdown.

    :rtype: dict
    """
    template = generate_generic_template(
        specifics=specifics, placeholders=placeholders,
        meta={'selected-item': '{{selected-item}}'}
    )
    template['action-type'] = 'select'
    return template


def generate_link_click_template(specifics: dict, placeholders: dict,
                                 **kwargs):
    """
    Generate a template for the click event of a link.

    :rtype: dict
    """
    template = generate_generic_template(
        specifics=specifics, placeholders=placeholders
    )
    template['action-type'] = 'click'
    return template


def generate_nav_button_click_template(specifics: dict, placeholders: dict,
                                       **kwargs):
    """
    Generate a template for the click event of a navigation button.

    :rtype: dict
    """
    template = generate_generic_template(
        specifics=specifics, placeholders=placeholders
    )
    template['action-type'] = 'click'
    return template


def generate_page_load_template(specifics: dict, placeholders: dict,
                                **kwargs):
    """
    Generate a template for the page load event of a page.

    :rtype: dict
    """
    template = generate_generic_template(
        specifics=specifics, placeholders=placeholders
    )
    template['action-type'] = 'load'
    return template


def generate_radio_buttons_check_template(specifics: dict, placeholders: dict,
                                          **kwargs):
    """
    Generate a template for the check event of a radio button.

    :rtype: dict
    """
    template = generate_generic_template(
        specifics=specifics, placeholders=placeholders
    )
    template['action-type'] = 'check'
    return template


def generate_submit_button_submit_template(specifics: dict, placeholders: dict,
                                           data: DataFrame, element_attrs: dict,
                                           **kwargs):
    """
    Generate a template for the submit event of a submit button.

    :rtype: dict
    """
    template = generate_generic_template(
        specifics=specifics, placeholders=placeholders
    )
    button_name = template['source']
    parent_name = '.'.join(button_name.split('.')[: -1])
    siblings = data.loc[
        (data['source'].str.startswith(parent_name)) &
        (data['source'].str.len() > len(parent_name)) &
        (data['source'] != button_name)
    ]
    for _, sibling in siblings.iterrows():
        element_path = sibling['source']
        element_name = element_path.split('.')[-1]
        element_type = sibling['ui-element']
        if element_type in element_attrs.keys():
            add_meta(template, element_name, '{{%s.%s}}' % (
                element_name,
                element_attrs[element_type]
            ))
    template['action-type'] = 'submit'
    return template


def generate_textbox_keystroke_template(specifics: dict, placeholders: dict,
                                        **kwargs):
    """
    Generate a template for the keystroke event of a textbox.

    :rtype: dict
    """
    template = generate_generic_template(
        specifics=specifics, placeholders=placeholders,
        meta={'key-id': '{{key-id}}'}
    )
    template['action-type'] = 'keystroke'
    return template



from typing import Dict, List

from ux.custom_types import SequenceFilter, SequenceFilterSet


class KPIConfig(object):

    def __init__(self, name: str, condition: SequenceFilter,
                 numerator_sets: List[str] = None, denominator_sets: List[str] = None,
                 filter_sets: Dict[str, SequenceFilterSet] = None):
        """
        Configuration for calculating a collection of KPI metrics which use the same filter condition in the numerator
        but will be carried out for different subsequences depending on numerator and denominator filters.
        Used in kpis.calculate_kpis_by_config.

        :param name: Name for the configuration object.
        :param condition: SequenceFilter that must be True for the sequence to be included in the numerator.
        :param numerator_sets: List of names of SequenceFilterSets to calculate numerators for each SequenceFilter.
        :param denominator_sets: List of names of SequenceFilterSets to calculate denominators for each SequenceFilter.
        :param filter_sets: Optional definitions of the numerator and denominator SequenceFilterSets
                            in format dict[FilterSetName, dict[FilterName, SequenceFilter]]
        """
        self.name = name
        self.condition = condition
        self.numerator_sets = numerator_sets or []
        self.denominator_sets = denominator_sets or []
        self.filter_sets = filter_sets or []



from collections import defaultdict
from datetime import datetime

from ux.interfaces.sequences.i_action_sequence import IActionSequence
from ux.interfaces.i_database_manager import IDatabaseManager


def find_location_history(manager: IDatabaseManager, start: datetime = None, end: datetime = None):
    """
    Find the history of each Location's appearance in the Database.

    :param manager: Instance of a class inheriting from IDatabaseManager.
    :param start: Optional start date-time to exclude older sessions.
    :param end: Optional end date-time to exclude newer sessions.
    :return: Dictionary mapping location ids to lists of session start times.
    :rtype: dict
    """
    history = defaultdict(list)
    for session in manager.sessions():
        session_start = session.start_time
        if (start and session_start < start) or (end and (session_start > end)):
            continue
        sequence: IActionSequence = manager.get_session_sequence(session_id=session.session_id)
        for location_id in sequence.location_ids():
            history[location_id].append(session_start)
    return dict(history)


def find_action_type_history(manager: IDatabaseManager, start: datetime = None, end: datetime = None):
    """
    Find the history of each Action Type's appearance in the Database.

    :param manager: Instance of a class inheriting from IDatabaseManager.
    :param start: Optional start date-time to exclude older sessions.
    :param end: Optional end date-time to exclude newer sessions.
    :return: Dictionary mapping location ids to lists of session start times.
    :rtype: dict
    """
    history = defaultdict(list)
    for session in manager.sessions():
        session_start = session.start_time
        if (start and session_start < start) or (end and (session_start > end)):
            continue
        sequence: IActionSequence = manager.get_session_sequence(session_id=session.session_id)
        for action_type in sequence.action_types():
            history[action_type].append(session_start)
    return dict(history)



from datetime import datetime

from ux.classes.actions.user_action import UserAction
from ux.classes.sequences.action_sequence import ActionSequence

a = UserAction(
    action_id='1',
    action_type='a',
    source_id='a',
    time_stamp=datetime.now(),
    user_id='1', session_id='1'
)
b = UserAction(
    action_id='1',
    action_type='b',
    source_id='b',
    time_stamp=datetime.now(),
    user_id='1', session_id='1'
)
c = UserAction(
    action_id='1',
    action_type='c',
    source_id='c',
    time_stamp=datetime.now(),
    user_id='1', session_id='1'
)
d = UserAction(
    action_id='1',
    action_type='d',
    source_id='d',
    time_stamp=datetime.now(),
    user_id='1', session_id='1'
)

seq = ActionSequence(
    user_actions=[a, b, c, a, a, b, a, c, d]
)

crop_seq = seq.crop(start=a.template(), end=c.template(), how='first')
print(crop_seq)
crop_seq = seq.crop(start=a.template(), end=c.template(), how='last')
print(crop_seq)
crop_seq = seq.crop(start=a.template(), end=b.template(), how='first')
print(crop_seq)
crop_seq = seq.crop(start=a.template(), end=b.template(), how='last')
print(crop_seq)
crop_seq = seq.crop(start=b.template(), end=c.template(), how='first')
print(crop_seq)
crop_seq = seq.crop(start=b.template(), end=c.template(), how='last')
print(crop_seq)
crop_seq = seq.crop(start=d.template(), end=a.template(), how='last')
print(crop_seq)



from typing import Dict

from ux.classes.wrappers.map_result import MapResult
from ux.interfaces.sequences.i_action_sequence import IActionSequence
from ux.interfaces.sequences.i_sequences_group_by import ISequencesGroupBy


class ISequences(object):

    @property
    def sequences(self):
        """
        Return a list of the individual Sequences in the collection.

        :rtype: List[IActionSequence]
        """
        raise NotImplementedError

    def filter(self, condition):
        """
        :rtype: ISequences
        """
        raise NotImplementedError

    def group_filter(self, filters):
        """
        :rtype: ISequencesGroupBy
        """
        raise NotImplementedError

    def count(self):
        """
        :rtype: int
        """
        raise NotImplementedError

    def copy(self):
        """
        :rtype: ISequences
        """
        raise NotImplementedError

    def intersection(self, other):
        """
        :rtype: ISequences
        """
        raise NotImplementedError

    @staticmethod
    def intersect_all(sequences):
        """
        :type sequences: List[ISequences]
        :rtype: ISequences
        """
        raise NotImplementedError

    def back_click_rates(self):
        """
        :rtype: Dict[IActionTemplate, float]
        """
        raise NotImplementedError

    def group_by(self, by):
        """
        :rtype: ISequencesGroupBy
        """
        raise NotImplementedError

    @property
    def durations(self):
        """
        :rtype: List[timedelta]
        """
        raise NotImplementedError

    def action_template_counts(self):
        """
        :rtype: Dict[IActionTemplate, int]
        """
        raise NotImplementedError

    def action_template_sequence_counts(self):
        """
        Return a total count of the number of ActionSequences containing each ActionTemplate in the collection.

        :rtype: Dict[IActionTemplate, int]
        """
        raise NotImplementedError

    def action_template_transition_counts(self):
        """
        Return counts of transitions between pairs of Actions from each Sequence in the collection.

        :return: Dictionary of {(from, to) => count}
        :rtype: Dict[tuple[IActionTemplate, IActionTemplate], int]
        """
        raise NotImplementedError

    def location_transition_counts(self, exclude=None):
        """
        Count the transitions from each location to each other location in actions in the given sequences.

        :return: Counter[Tuple[from, to], count]
        :rtype: Counter[Tuple[str, str], int]
        """
        raise NotImplementedError

    def dwell_times(self, sum_by_location: bool, sum_by_sequence: bool):
        """
        Return the amount of time spent by the user at each location.

        :param sum_by_location: Whether to sum the durations of time spent at each location or keep as a list.
        :param sum_by_sequence: Whether to sum the durations of time spent at each location in each sequence or keep as a list.
        :rtype: Dict[str, Union[timedelta, List[timedelta]]]
        """
        raise NotImplementedError

    def map(self, mapper):
        """
        :rtype: MapResult
        """
        raise NotImplementedError

    def sort(self, by: str, ascending: bool = True):
        """
        :rtype: ISequences
        """
        raise NotImplementedError

    def __getitem__(self, item):

        raise NotImplementedError

    def __len__(self):

        raise NotImplementedError

    def __contains__(self, item: IActionSequence):

        raise NotImplementedError

    def __iter__(self):
        """
        :rtype: IActionSequence
        """
        raise NotImplementedError

    def __add__(self, other):
        """
        :rtype: ISequences
        """
        raise NotImplementedError

    def __sub__(self, other):
        """
        :rtype: ISequences
        """
        raise NotImplementedError



class ISequencesGroupBy(object):

    def count(self):

        raise NotImplementedError

    def items(self):

        raise NotImplementedError

    def keys(self):
        """
        :rtype: dict_keys
        """
        raise NotImplementedError

    def values(self):
        """
        :rtype: dict_values
        """
        raise NotImplementedError

    @property
    def names(self):
        """
        :rtype: List[str]
        """
        raise NotImplementedError

    def map(self, mapper):

        raise NotImplementedError

    def agg(self, agg_funcs: dict):
        """
        :param agg_funcs: dict mapping attributes to one or more aggregation functions e.g. duration -> np.median
        """
        raise NotImplementedError

    def filter(self, condition):
        """
        Return a new Sequences containing only the sequences matching the `condition` in each group.

        :param condition: lambda(sequence) that returns True to include a sequence.
        :rtype: ISequencesGroupBy
        """
        raise NotImplementedError

    def group_filter(self, filters, group_name=None):
        """
        Return a new SequencesGroupBy keyed by the filter name with values matching each filter, applied in parallel.

        :param filters: Dictionary of filters to apply.
        :param group_name: Name to identify the filter group.
        :rtype: SequencesGroupBy
        """
        raise NotImplementedError

    def __getitem__(self, item):
        """
        :rtype: ISequences
        """
        raise NotImplementedError



from typing import Callable, Dict

from ux.custom_types import SequenceFilter, ActionFilter
from ux.interfaces.actions.i_user_action import IUserAction
from ux.interfaces.sequences.i_action_sequence import IActionSequence


class CountConfig(object):

    def __init__(self, name: str,
                 sequence_condition: SequenceFilter = None,
                 sequence_split_by: Callable[[IActionSequence], Dict[str, int]] = None,
                 action_condition: ActionFilter = None,
                 action_split_by: Callable[[IUserAction], Dict[str, int]] = None):
        """
        Configuration class for batch calculation of count metrics.

        - If counting sequences, must pass `sequence_condition` and `sequence_split_by`.
        - If counting actions within sequences, must also pass `action_condition` and `action_split_by`
        - set `sequence_condition` or `action_condition` to `lambda x: True` to include all sequences or actions

        :param name: Name of the count metric.
        :param sequence_condition: lambda(sequence) that must return True for a sequence to be included in the count.
        :param sequence_split_by:  lambda(sequence) that splits sequence counts into dict[split_name, count]
        :param action_condition: lambda(action) that must return True for an action to be included in the count.
        :param action_split_by: lambda(action) that splits action counts into dict[split_name, count]
        """
        if action_condition is not None:
            assert sequence_condition is not None
        self.name = name
        self.sequence_condition = sequence_condition
        self.sequence_split_by = sequence_split_by
        self.action_condition = action_condition
        self.action_split_by = action_split_by

    def __repr__(self):

        args = ['name', 'sequence_condition', 'sequence_split_by', 'action_condition', 'action_split_by']
        return 'CountConfig({})'.format(', '.join(['{}={}'.format(arg, getattr(self, arg)) for arg in args]))



from datetime import date, datetime, timedelta


def monday_on_or_before(the_date: date):
    """
    Return the date of the Monday in the week of the given date.

    :param the_date: The date to find the date of the Monday for.
    :rtype: date
    """
    return the_date - timedelta(days=the_date.weekday())


def monday_on_or_after(the_date: date):
    """
    Return the date of the Monday in the week of the given date.

    :param the_date: The date to find the date of the Monday for.
    :rtype: date
    """
    return the_date + timedelta(days=-the_date.weekday(), weeks=the_date.weekday() > 1)


def date_to_datetime(the_date: date):
    """
    Convert the given date to a datetime instance.

    :param the_date: The date to convert.
    :rtype: datetime
    """
    return datetime(
        the_date.year, the_date.month, the_date.day
    )



class ISession(object):

    @property
    def session_id(self):
        """
        :rtype: str
        """
        raise NotImplementedError

    @property
    def user_id(self):
        """
        :rtype: str
        """
        raise NotImplementedError

    @property
    def start_time(self):
        """
        :rtype: datetime
        """
        raise NotImplementedError

    @property
    def end_time(self):
        """
        :rtype: datetime
        """
        raise NotImplementedError



from matplotlib.axes import Axes
from matplotlib.dates import DateFormatter, MinuteLocator, HourLocator, DayLocator, SecondLocator, MonthLocator, \
    YearLocator
from matplotlib.patches import Circle, FancyArrowPatch, ConnectionStyle, ArrowStyle
from numpy.ma import arange
from pandas import DataFrame
from seaborn import heatmap
from typing import Dict, List

from ux.interfaces.sequences.i_action_sequence import IActionSequence
from ux.plots.helpers import new_axes, point_distance, circle_edge, get_color
from ux.utils.transitions import create_transition_matrix


def plot_transition_matrix(transitions, get_name=None,
                           order_by='from', exclude=None,
                           ax: Axes = None, heatmap_kws: dict = None):
    """
    Plot a state transition matrix from the given transition counts or probabilities.

    :param transitions: Dictionary of transitions and their counts or probabilities.
    :type transitions: Dict[Tuple[object, object], Union[float, int]]
    :param get_name: Optional lambda function to call to convert states to labels.
    :type get_name: Callable[[IUserAction], str]
    :param order_by: Order labels by descending count of `from` or `to`, or pass a list to set order explicitly.
    :type order_by: Union[str, List[str]]
    :param exclude: Optional list of labels to exclude from the plots.
    :type exclude: Union[str, List[str]]
    :param heatmap_kws: Keyword args and values for seaborn's heatmap function.
    :param ax: Optional matplotlib axes to plot on.
    :rtype: Axes
    """
    matrix = create_transition_matrix(transitions=transitions, get_name=get_name,
                                      order_by=order_by, exclude=exclude)
    ax = ax or new_axes()
    if heatmap_kws is None:
        heatmap_kws = {}
    heatmap(matrix, ax=ax, **heatmap_kws, annot=True, fmt='0.0f')
    ax.set_xticklabels(matrix.columns.tolist())
    ax.set_yticklabels(matrix.columns.tolist())
    ax.invert_yaxis()
    ax.figure.tight_layout()
    return ax


def plot_markov_chain(transitions, get_location, get_name=None,
                      state_color=None, transition_color=None, arc_scale: float = 0.1,
                      text_kws: dict = None, circle_kws: dict = None, arrowstyle_kws: dict = None,
                      ax: Axes = None):
    """
    Plot a diagram of the Markov Chain corresponding to the given transitions between states.

    :param transitions: List of transitions and their counts or probabilities.
    :type transitions: Dict[Tuple[object, object], Union[float, int]]
    :param get_location: Lambda function to call to get state plot locations.
    :type get_location: Callable[[IUserAction], (float, float)]
    :param get_name: Optional lambda function to call to convert states to labels.
    :type get_name: Callable[[IUserAction], str]
    :param state_color: string or callable(state) to generate color for each state.
    :type state_color: Union[str, Callable]
    :param transition_color: string or callable(state) to generate color for each transition.
    :type transition_color: Union[str, Callable]
    :param arc_scale: this is multiplied by the distance between states to get the radius of the connecting arc
    :param text_kws: args to pass to `ax.text` for the state names
    :param circle_kws: args to pass to `matplotlib.patches.Circle` for the states
    :param arrowstyle_kws: args to pass to `matplotlib.patches.FancyArrowPatch(arrowstyle)` for the transitions
    :param ax: Optional matplotlib axes to plot on.
    :rtype: Axes
    """
    ax = ax or new_axes()
    circle_kws = circle_kws or {'radius': 0.35}
    text_kws = text_kws or {'ha': 'center', 'va': 'center'}
    get_name = get_name or str
    # get unique states
    states = set()
    for from_state, to_state in transitions.keys():
        states.update([from_state, to_state])
    # plot state circles
    for state in states:
        center = get_location(state)
        color = get_color(state_color, state, default='green')
        ax.add_patch(Circle(
            xy=center, color=color, **circle_kws
        ))
        ax.plot(*center, c='yellow')
    # plot transition arrows
    arrowstyle_kws = arrowstyle_kws or dict(stylename='Fancy', head_length=10, head_width=5, tail_width=2)
    for (from_state, to_state), number in transitions.items():
        from_center = get_location(from_state)
        to_center = get_location(to_state)
        distance = point_distance(from_center, to_center)
        circle_radius = circle_kws['radius']
        if distance > 0:
            from_edge = circle_edge(from_center, to_center, circle_radius, 15)
            to_edge = circle_edge(to_center, from_center, circle_radius, -15)
            color = get_color(transition_color, from_state, default='red')
            ax.add_patch(FancyArrowPatch(
                posA=from_edge, posB=to_edge,
                connectionstyle=ConnectionStyle(stylename='Arc3', rad=arc_scale*distance),
                arrowstyle=ArrowStyle(**arrowstyle_kws),
                linewidth=0, color=color
            ))
    # plot state labels
    for state in states:
        center = get_location(state)
        ax.text(*center, s=get_name(state), **text_kws)


def plot_sequence_diagram(sequence: IActionSequence, locations: List[str], max_grid_lines: int = 50):
    """
    :rtype: Axes
    """
    data = sequence.map({
        'location': lambda act: act.source_id,
        'action-type': lambda act: act.action_type,
        'time_stamp': lambda act: act.time_stamp
    }).to_frame(wide=True)
    data['y'] = [
        len(locations) - locations.index(location) - 1
        if location in locations
        else len(locations)
        for location in data['location']
    ]
    data['label'] = [location if location not in locations else '' for location in data['location']]
    ax = new_axes()
    ax.plot_date(x=data['time_stamp'], y=data['y'],
                 marker='o', linestyle='-')
    for _, row in data.iterrows():
        ax.text(x=row['time_stamp'], y=row['y'], s=row['label'],
                ha='center', va='bottom', rotation=45)
    formatter = DateFormatter("%H:%M:%S")
    ax.xaxis.set_major_formatter(formatter)
    ax.set(
        yticks=arange(len(locations) + 1),
        yticklabels=(['other'] + locations)[:: -1]
    )
    x_min, x_max = ax.get_xlim()
    min_spacing = (x_max - x_min) / max_grid_lines
    locators = [
        (1 / (24 * 60 * 60 * 12), SecondLocator(interval=1), SecondLocator(interval=1)),  # 5ms
        (1 / (24 * 60 * 60), SecondLocator(interval=1), SecondLocator(interval=5)),  # 1s
        (1 / (24 * 60 * 12), SecondLocator(interval=5), SecondLocator(interval=30)),  # 5s
        (1 / (24 * 60), MinuteLocator(interval=1), MinuteLocator(interval=5)),  # 1m
        (1 / (24 * 12), MinuteLocator(interval=5), HourLocator(interval=1)),  # 5m
        (1 / 24, HourLocator(interval=5), HourLocator(interval=6)),  # 1h
        (1, DayLocator(interval=1), DayLocator(interval=7)),  # 1d
        (7, DayLocator(interval=7), DayLocator(interval=30)),  # 1w
        (30, MonthLocator(interval=1), YearLocator()),  # 1M
    ]
    minor_locator = locators[-1][1]
    major_locator = locators[-1][2]
    for locator in locators[::-1]:
        if min_spacing < locator[0]:
            minor_locator = locator[1]
            major_locator = locator[2]
    ax.xaxis.set_minor_locator(minor_locator)
    ax.xaxis.set_major_locator(major_locator)
    ax.grid(which='minor', lw=0.5)
    ax.grid(which='major', lw=1)
    return ax



from pandas import DataFrame, Series, MultiIndex, Index, concat
from typing import List, Union, Iterable


def _str_or_non_iterable(val):

    return not isinstance(val, Iterable) or isinstance(val, str)


class MapResult(object):

    def __init__(self, data: dict,
                 key_names: Union[str, List[str]] = 'map',
                 value_names: Union[str, List[str]] = 'result'):

        if isinstance(key_names, str):
            key_names = [key_names]
        if isinstance(value_names, str):
            value_names = [value_names]

        # check key names
        if all(_str_or_non_iterable(k) for k in list(data.keys())):
            assert len(key_names) == 1, 'Length of index names must be 1 when keys are not Iterables'
        else:
            assert len(key_names) == len(list(data.keys())[0]), 'Keys must have same length as index_names'

        # check value names
        assert len(value_names) == 1, 'value_names must be length 1'

        self._data: dict = data
        self._index_names = key_names
        self._data_names = value_names

        self._first_key = list(data.keys())[0]
        self._first_value = list(data.values())[0]

        for attribute in self._data.keys():
            try:
                setattr(self, attribute, self._data[attribute])
            except:
                pass

    @property
    def index_names(self):
        return self._index_names

    @property
    def data_names(self):
        return self._data_names

    def to_series(self):
        """
        :rtype: Series
        """
        if _str_or_non_iterable(self._first_key):
            if _str_or_non_iterable(self._first_value):
                # e.g. {'a': 1, 'b': 2, 'c': 3}
                series = Series(self._data, name=self.data_names[0])
                series.index.name = self.index_names[0]
            else:
                # e.g. {'a': [1, 2], 'b': [3, 4, 5]}
                series = concat([
                    Series(
                        index=Index([key] * len(values), name=self.index_names[0]),
                        data=values, name=self.data_names[0]
                    )
                    for key, values in self._data.items()
                ])
        else:
            if _str_or_non_iterable(self._first_value):
                # e.g. {('a', 'b'): 1, ('c', 'd'): 2, ('e', 'f'): 3}
                series = Series(self._data, name=self.data_names[0])
                series.index.names = self.index_names
            else:
                # e.g. {('a', 'b'): [1, 2], ('c', 'd'): [3, 4, 5]}
                series = concat([
                    Series(
                        index=MultiIndex.from_tuples([key] * len(values), names=self.index_names),
                        data=values, name=self.data_names[0]
                    )
                    for key, values in self._data.items()
                ])

        return series

    def to_dict(self):
        """
        :rtype: dict
        """
        return self._data

    def to_frame(self, wide: bool = False):
        """
        :rtype: DataFrame
        """
        if not wide:
            return self.to_series().to_frame().reset_index()
        else:
            try:
                return DataFrame(self._data)
            except:
                return self.to_series().to_frame().reset_index()

    def to_list(self):
        """
        :rtype: list
        """
        return list(self.to_series().tolist())

    def to_tuples(self):
        """
        :rtype: List[tuple]
        """
        return list(self.to_frame().to_records())

    def items(self):

        return self._data.items()

    def keys(self):

        return self._data.keys()

    def values(self):

        return self._data.values()

    def __getitem__(self, item):

        return self._data[item]

    def __iter__(self):

        return self._data.__iter__()

    def __repr__(self):

        return 'MapResult({})'.format(self._data.__repr__())



from typing import Callable

from ux.interfaces.sequences.i_action_sequence import IActionSequence
from ux.interfaces.tasks.i_task import ITask


def unordered_task_completion_rate(task: ITask, action_sequence: IActionSequence):
    """
    Calculate the Task completion from a Sequence of User Actions.
    Does not require the Actions to be completed in order.

    :param task: The Task to measure completion rate for.
    :param action_sequence: The Sequence of user actions.
    :rtype: float
    """
    # calculate sum of task unique action template weights
    task_templates = task.action_templates
    task_template_set = set(task_templates)
    task_weight = sum([action_template.weighting
                       for action_template in task.action_templates])
    # calculate sum of weights of action templates in task and sequence
    sequence_template_set = set(action_sequence.action_templates())
    overlaps = task_template_set.intersection(sequence_template_set)
    overlap_weight = sum([
        action_template.weighting
        for action_template in overlaps
    ])
    return overlap_weight / task_weight


def ordered_task_completion_rate(task: ITask, action_sequence: IActionSequence):
    """
    Calculate the Task completion from a sequence of User Actions.
    Requires that the Actions are completed in the order specified in the Task.

    :param task: The Task to measure completion rate for.
    :param action_sequence: The Sequence of User Actions
    :rtype: float
    """
    # calculate sum of task action template weights
    task_templates = task.action_templates
    task_weight = sum([action_template.weighting
                       for action_template in task.action_templates])
    num_templates = len(task_templates)
    sequence_templates = action_sequence.action_templates()
    # find weights of action templates in sequence in order
    found = True
    num_found = 0
    sequence_weight = 0.0
    while found and num_found < num_templates:
        search_template = task_templates[num_found]
        try:
            sequence_index = sequence_templates.index(search_template)
            num_found += 1
            sequence_weight += search_template.weighting
            sequence_templates = sequence_templates[sequence_index + 1:]
        except:
            found = False

    return sequence_weight / task_weight


def binary_task_success(task: ITask, action_sequence: IActionSequence,
                        success_func: Callable[[ITask, IActionSequence], bool]):
    """
    Calculate the binary task success for each sequence.

    :param task: The Task to assess success against.
    :param action_sequence: ActionSequence to assess success of.
    :param success_func: Callable to use to assess success.
    :return: List of False for fail or True for pass
    :rtype: bool
    """
    return success_func(task, action_sequence)



from datetime import datetime
from matplotlib.axes import Axes

from ux.interfaces.i_database_manager import IDatabaseManager
from ux.plots.helpers import new_axes
from ux.utils.versioning import find_location_history, find_action_type_history


def plot_history(manager: IDatabaseManager, start: datetime = None, end: datetime = None, ax: Axes = None,
                 history_type: str = 'location'):
    """
    Plot the history of each Location's appearance in the set of logs in Manager.

    :param manager: Child of IDatabaseManager containing the logs to use.
    :param start: Optional start date-time to exclude older sessions.
    :param end: Optional end date-time to exclude newer sessions.
    :param ax: Optional matplotlib axes to plot on.
    :rtype: Axes
    """
    ax = ax or new_axes()
    if history_type == 'location':
        find_history = find_location_history
    elif history_type == 'action-type':
        find_history = find_action_type_history
    else:
        raise ValueError("history_type must be 'location' or 'action-type'")
    history = find_history(manager=manager, start=start, end=end)
    locations = list(history.keys())
    # sort locations by first session
    locations = sorted(locations, key=lambda loc: history[loc][0])
    # plot histories
    for l, location in enumerate(locations):
        color = 'C{}'.format(l % 10)
        ax.scatter(
            x=history[location],
            y=[l] * len(history[location]),
            marker='x', s=5, c=color
        )
        x = [history[location][0], history[location][-1]]
        y = [l, l]
        ax.plot(x, y, c=color, alpha=0.1)
    # format axes
    ax.set_yticks(range(1, len(locations)))
    ax.set_yticklabels(locations)
    if start:
        x_lim = ax.get_xlim()
        ax.set_xlim(start, x_lim[1])
    if end:
        x_lim = ax.get_xlim()
        ax.set_xlim(x_lim[0], end)



from types import FunctionType
from typing import Union


def get_method_name(method: Union[str, FunctionType]):
    """
    :rtype: str
    """
    if isinstance(method, str):
        return method
    elif isinstance(method, FunctionType):
        name = method.__name__
        if name != '<lambda>':
            return name
        else:
            return str(method)



from types import FunctionType

from ux.interfaces.actions.i_action_template import IActionTemplate


def create_action_template_condition(value):
    """
    :type value: Union[IActionTemplate, FunctionType]
    :rtype: FunctionType
    """
    def action_template_condition(template: IActionTemplate):
        if template == value:
            return True
        else:
            return False

    if isinstance(value, IActionTemplate):
        return action_template_condition
    elif isinstance(value, FunctionType):
        return value
    else:
        raise TypeError('expected IActionTemplate or FunctionType')



from typing import Iterable

from pandas import Series
from statsmodels.stats.proportion import proportion_confint


def binary_task_success_rate(results: Iterable[int],
                             alpha: float = 0.05, method: str = 'normal'):
    """
    Calculate the binary task success rate from a list of pass or fail task results for a given user.

    See http://www.statsmodels.org/devel/generated/statsmodels.stats.proportion.proportion_confint.html

    :param results: list of pass (1/True) / fail (0/False) results
    :param alpha: significance level
    :param method: method to use for confidence interval
    :rtype: SuccessRate
    """
    s: Series = Series(results).astype(int)
    mean = s.mean()
    confidence_interval = proportion_confint(
        count=s.sum(), nobs=s.shape[0],
        alpha=alpha, method=method

    )

    return mean, confidence_interval


class IUser(object):

    @property
    def user_id(self):
        """
        :rtype: str
        """
        raise NotImplementedError

    @property
    def session_ids(self):
        """
        :rtype: List[str]
        """
        raise NotImplementedError

    @property
    def action_ids(self):
        """
        :rtype: List[str]
        """
        raise NotImplementedError

    def add_action_id(self, action_id):
        raise NotImplementedError

    def add_session_id(self, session_id):
        raise NotImplementedError



from typing import Iterable

from numpy import mean, std
from scipy.stats import norm, expon


def normal_confidence_interval(data: Iterable, confidence: float = 0.95):
    """
    Compute confidence interval for array or list which is assumed to be normally distributed.

    :param data: list or array
    :param confidence: confidence level between 0 and 1
    :rtype: Tuple[float, float]
    """
    mu, sigma = mean(data), std(data, ddof=1)
    lower, upper = norm.interval(confidence, loc=mu, scale=sigma)
    error_lower = mu - lower
    error_upper = upper - mu
    return error_lower, error_upper


def exponential_confidence_interval(data: Iterable, confidence: float = 0.95):
    """
    Compute confidence interval for array or list which is assumed to be exponentially distributed.

    :param data: list or array
    :param confidence: confidence level between 0 and 1
    :rtype: Tuple[float, float]
    """
    shift, mu = expon.fit(data)
    interval = expon.interval(alpha=1-confidence, loc=shift, scale=mu)
    return mu - interval[0], interval[1] - mu


class IActionTemplate(object):

    @property
    def action_type(self):
        """
        Return the type of the ActionTemplate.

        :rtype: str
        """
        raise NotImplementedError

    @property
    def source_id(self):
        """
        Return the id of the Location where the Action would be taken.

        :rtype: str
        """
        raise NotImplementedError

    @property
    def target_id(self):
        """
        Return the id of the Location where the Action would go to.

        :rtype: str
        """
        raise NotImplementedError

    @property
    def weighting(self):
        """
        Return the weighting for the Action for various calculations.

        :rtype: float
        """
        raise NotImplementedError

    def reversed(self, action_type: str = None):
        """
        Return a reversed version of the template (switch the source and target).

        :param action_type: Optional action type. Leave as None to use the existing action type.
        :rtype: IActionTemplate
        """
        raise NotImplementedError

    def to_dict(self):
        """
        :rtype: dict
        """
        raise NotImplementedError



from datetime import datetime
import unittest
from unittest import TestCase

from ux.calcs.object_calcs.task_success import ordered_task_completion_rate, unordered_task_completion_rate
from ux.classes.sequences.action_sequence import ActionSequence
from ux.classes.actions.action_template import ActionTemplate
from ux.classes.tasks.task import Task
from ux.classes.actions.user_action import UserAction


class TestTaskSuccess(TestCase):

    def setUp(self):

        self.test_task = Task(
            name='OrderedTask',
            action_templates=[
                ActionTemplate(
                    action_type='test-action',
                    source_id='location-{}'.format(a + 1),
                ) for a in range(5)
            ]
        )
        self.ordered_test_sequence = ActionSequence(
            user_actions=[
                UserAction(
                    action_id='action-{}'.format(i),
                    action_type='test-action',
                    source_id='location-{}'.format(a),
                    time_stamp=datetime.now(), user_id='user_1', session_id='session_1'
                ) for i, a in enumerate([1, 2, 3, 2, 3, 4, 5])
            ]
        )
        self.unordered_test_sequence = ActionSequence(
            user_actions=[
                UserAction(
                    action_id='action-{}'.format(i),
                    action_type='test-action',
                    source_id='location-{}'.format(a),
                    time_stamp=datetime.now(), user_id='user_1', session_id='session_1'
                ) for i, a in enumerate([1, 2, 3, 2, 5, 4])
            ]
        )

    def test_ordered_task_completion_rate_1(self):

        self.assertEqual(
            1,
            ordered_task_completion_rate(
                task=self.test_task,
                action_sequence=self.ordered_test_sequence
            )
        )

    def test_ordered_task_completion_rate_2(self):

        self.assertEqual(
            0.8,
            ordered_task_completion_rate(
                task=self.test_task,
                action_sequence=self.unordered_test_sequence
            )
        )

    def test_unordered_task_completion_rate_1(self):

        self.assertEqual(
            1,
            unordered_task_completion_rate(
                task=self.test_task,
                action_sequence=self.ordered_test_sequence
            )
        )

    def test_unordered_task_completion_rate_2(self):

        self.assertEqual(
            1,
            unordered_task_completion_rate(
                task=self.test_task,
                action_sequence=self.unordered_test_sequence
            )
        )


if __name__ == '__main__':

    unittest.main()



import matplotlib.pyplot as plt

from ux.plots.usability.errors import plot_percentage_task_errors


def percentage_task_errors():
    """
    Emulate the plot from Figure 4.8 of "Measuring the User Experience".
    """
    # generate example data
    condition_result = {
        'Control': [1, 0, 0, 0, 0, 0, 1, 1, 0, 0, 1, 1],
        'Condition 1': [1, 1, 1, 1, 0, 1, 0, 0, 1, 1, 0, 1],
        'Condition 2': [0, 0, 1, 1, 0, 0, 1, 0, 0, 1, 1, 0],
        'Condition 3': [1, 1, 1, 1, 1, 0, 1, 1, 1, 1, 1, 1],
        'Condition 4': [0, 0, 1, 1, 1, 1, 1, 1, 0, 1, 1, 1],
        'Condition 5': [0, 1, 1, 1, 1, 1, 1, 0, 1, 1, 1, 1],
        'Condition 6': [0, 0, 1, 1, 0, 0, 1, 1, 0, 1, 1, 1],
        'Condition 7': [0, 1, 1, 1, 0, 1, 1, 0, 1, 0, 1, 1],
        'Condition 8': [0, 1, 0, 1, 0, 1, 1, 0, 1, 0, 1, 1]
    }
    # plot
    plot_percentage_task_errors(condition_result)
    plt.show()


if __name__ == '__main__':

    percentage_task_errors()



from ux.interfaces.actions.i_action_template import IActionTemplate


class ActionTemplate(IActionTemplate):
    """
    Represents a Template for a single Action that could be taken by a User.
    """
    def __init__(self, action_type: str, source_id: str, target_id: str = None,
                 weighting: float = 1):
        """
        Create a new ActionTemplate.

        :param action_type: A type for the Action (used in matching)
        :param source_id: The id of the Location where the Action would be taken (used in matching).
        :param target_id: Optional id of the Location where the Action would go to (used in matching).
        :param weighting: Optional weighting value to use in various calculations.
        """
        self._action_type = action_type
        self._source_id = source_id
        self._target_id = target_id
        self._weighting = weighting

    @property
    def action_type(self):
        """
        Return the type of the ActionTemplate.

        :rtype: str
        """
        return self._action_type

    @property
    def source_id(self):
        """
        Return the id of the Location where the Action would be taken.

        :rtype: str
        """
        return self._source_id

    @property
    def target_id(self):
        """
        Return the id of the Location where the Action would go to.

        :rtype: str
        """
        return self._target_id

    @property
    def weighting(self):
        """
        Return the weighting for the Action for various calculations.

        :rtype: float
        """
        return self._weighting

    def reversed(self, action_type: str = None):
        """
        Return a reversed version of the template (switch the source and target).

        :param action_type: Optional action type. Leave as None to use the existing action type.
        :rtype: IActionTemplate
        """
        return ActionTemplate(
            source_id=self.target_id,
            target_id=self.source_id,
            action_type=action_type or self.action_type
        )

    def to_dict(self):
        """
        :rtype: dict
        """
        return {
            'action_type': self._action_type,
            'source_id': self._source_id,
            'target_id': self._target_id
        }

    def __eq__(self, other):
        """
        :type other: ActionTemplate
        """
        return (
            (self._action_type == other._action_type or '*' in (self._action_type, other._action_type)) and
            (self._source_id == other._source_id or '*' in (self._source_id, other._source_id)) and
            (self._target_id == other._target_id or '*' in (self._target_id, other._target_id))
        )

    def __repr__(self):

        return 'ActionTemplate({}: {}{}{})'.format(
            self._action_type, self._source_id,
            ' → ' if self._target_id else '',
            self._target_id if self._target_id else ''
        )

    def __hash__(self):

        return hash(tuple(sorted(self.to_dict().items())))



from typing import List, Set

from ux.interfaces.actions.i_action_template import IActionTemplate
from ux.interfaces.actions.i_user_action import IUserAction
from ux.classes.wrappers.map_result import MapResult


class IActionSequence(object):

    @property
    def user_actions(self):
        """
        :rtype: List[IUserAction]
        """
        raise NotImplementedError

    def action_templates(self):
        """
        :rtype: List[IActionTemplate]
        """
        raise NotImplementedError

    def action_template_set(self):
        """
        Return a set of ActionTemplates representing the UserActions in the ActionSequence.

        :rtype: Set[IActionTemplate]
        """
        raise NotImplementedError

    def action_template_counts(self):
        """
        Return a count of each ActionTemplate representing one or more UserActions in the ActionSequence.

        :rtype: Dict[IActionTemplate, int]
        """
        raise NotImplementedError

    @property
    def meta(self):
        """
        :rtype: dict
        """
        raise NotImplementedError

    @property
    def duration(self):
        """
        :rtype: timedelta
        """
        raise NotImplementedError

    @property
    def start(self):
        """
        :rtype: datetime
        """
        raise NotImplementedError

    @property
    def end(self):
        """
        :rtype: datetime
        """
        raise NotImplementedError

    @property
    def user_id(self):
        """
        :rtype: str
        """
        raise NotImplementedError

    @property
    def session_id(self):
        """
        :rtype: str
        """
        raise NotImplementedError

    def map(self, mapper):
        """
        :rtype: MapResult
        """
        raise NotImplementedError

    def unordered_completion_rate(self, task):
        """
        :type task: ITask
        :rtype: float
        """
        raise NotImplementedError

    def ordered_completion_rate(self, task):
        """
        :type task: ITask
        :rtype: float
        """
        raise NotImplementedError

    def intersects_task(self, task):
        """
        :type task: ITask
        :rtype: bool
        """
        raise NotImplementedError

    def action_types(self):
        """
        :rtype: Set[str]
        """
        raise NotImplementedError

    def back_click_rates(self):
        """
        :rtype: Dict[IActionTemplate, float]
        """
        raise NotImplementedError

    def dwell_times(self, sum_by_location: bool):
        """
        Return the amount of time spent by the user at each location.

        :param sum_by_location: Whether to sum the durations of time spent at each location or keep as a list.
        :rtype: Dict[str, Union[timedelta, List[timedelta]]]
        """
        raise NotImplementedError

    def location_ids(self):
        """
        :rtype: Set[str]
        """
        raise NotImplementedError

    def contains_location_id(self, location_id: str):
        """
        Determine whether the location was visited in the sequence.

        :rtype: bool
        """
        raise NotImplementedError

    def find_all(self, template: IActionTemplate):
        """
        Return a list of all the actions matching the given action template.
        Returns an empty list if the template is not matched.

        :param template: The ActionTemplate to match against.
        :rtype: list[IUserAction]
        """
        raise NotImplementedError

    def find_first(self, template: IActionTemplate):
        """
        Return the first action matching the given action template. Returns None if the template is not matched.

        :param template: The ActionTemplate to match against.
        :rtype: IUserAction
        """
        raise NotImplementedError

    def find_last(self, template: IActionTemplate):
        """
        Return the first action matching the given action template. Returns None if the template is not matched.

        :param template: The ActionTemplate to match against.
        :rtype: IUserAction
        """
        raise NotImplementedError

    def __getitem__(self, item):
        """
        :rtype: IUserAction
        """
        raise NotImplementedError

    def __len__(self):
        """
        :rtype: int
        """
        raise NotImplementedError

    def __contains__(self, item):
        """
        :rtype: bool
        """
        raise NotImplementedError

    def __iter__(self):
        """
        :rtype: IUserAction
        """
        raise NotImplementedError



from typing import Tuple

from ux.interfaces.sequences.i_action_sequence import IActionSequence
from ux.interfaces.tasks.i_task import ITask


def task_extents(action_sequence: IActionSequence, task: ITask):
    """
    Return the start and end index of the actions in the sequence that are part of the task.

    :return: Index of the first and last User Actions in the Sequence that are also in the Task.
    :rtype: Tuple[int, int]
    """
    i_first = 0
    i_last = len(action_sequence) - 1
    sequence_templates = action_sequence.action_templates()
    while sequence_templates[i_first] not in task.action_templates:
        i_first += 1
    while sequence_templates[i_last] not in task.action_templates:
        i_last -= 1
    return i_first, i_last


def time_on_task(task: ITask, action_sequence: IActionSequence):
    """
    Measure the time from the start of the first action in the task to the end of the last action.

    :param task: Task definition.
    :param action_sequence: Sequence of actions
    """
    i_first, i_last = task_extents(action_sequence=action_sequence, task=task)
    first_action = action_sequence.user_actions[i_first]
    last_action = action_sequence.user_actions[i_last]
    return last_action.time_stamp - first_action.time_stamp



from typing import List


class IDatabaseManager(object):
    """
    An interface or abstract base class that should be implemented for each new backend.
    """
    def session(self, session_id: str):
        """
        Return the Session with the given id.

        :rtype: ISession
        """
        raise NotImplementedError

    def sessions(self):
        """
        Return all the Sessions in the database.

        :rtype: List[ISession]
        """
        raise NotImplementedError

    def user(self, user_id: str):
        """
        :rtype: IUser
        """
        raise NotImplementedError

    def users(self):
        """
        Return all the Users in the database.

        :rtype: List[IUser]
        """
        raise NotImplementedError

    def user_action(self, action_id: str):
        """
        Return the UserAction with the given id.

        :rtype: IUserAction
        """
        raise NotImplementedError

    def user_actions(self, user_id: str):
        """
        Return all the Actions taken by the User with the given id in the database.

        :rtype: List[IUserAction]
        """
        raise NotImplementedError

    def location(self, location_id: str):
        """
        Return the Location with the given id.

        :rtype: ILocation
        """
        raise NotImplementedError

    def locations(self):
        """
        Return all the Locations in the database.

        :rtype: List[ILocation]
        """
        raise NotImplementedError

    def action_types(self):
        """
        Return all the Action Types embedded in UserActions.

        :rtype: List[str]
        """
        raise NotImplementedError

    def get_session_sequence(self, session_id):
        """
        Return an ActionSequence constructed from all UserActions in the Session with the given id.

        :rtype: IActionSequence
        """
        raise NotImplementedError



from setuptools import setup, find_packages

setup(
    name='ux',
    version='0.1.13',
    packages=find_packages(),
    url='https://github.com/vahndi/ux',
    license='MIT',
    author='vahndi.minah',
    author_email='',
    description='A Python package for measuring and analyzing User Experience',
    install_requires=[
        'matplotlib',
        'numba',
        'numpy',
        'pandas',
        'scipy',
        'seaborn',
        'statsmodels'
    ]
)
